import uuid as _uuid

import six as _six
from google.protobuf.json_format import MessageToDict as _MessageToDict

from flytekit.common import constants as _constants
from flytekit.common import interface as _interface
from flytekit.common import nodes as _nodes
from flytekit.common.exceptions import scopes as _exception_scopes
from flytekit.common.exceptions.user import FlyteTypeException as _FlyteTypeException
from flytekit.common.exceptions.user import FlyteValueException as _FlyteValueException
from flytekit.common.tasks import output as _task_output
from flytekit.common.tasks import sdk_runnable as _sdk_runnable
from flytekit.common.tasks import task as _base_task
from flytekit.common.types import helpers as _type_helpers
from flytekit.models import dynamic_job as _dynamic_job
from flytekit.models import interface as _interface_model
from flytekit.models import literals as _literal_models
from flytekit.models import qubole as _qubole
from flytekit.models.core import workflow as _workflow_model

ALLOWED_TAGS_COUNT = int(6)
MAX_TAG_LENGTH = int(20)


class SdkHiveTask(_sdk_runnable.SdkRunnableTask):
    """
    This class includes the additional logic for building a task that executes as a batch hive task.
    """

    def __init__(
        self,
        task_function,
        task_type,
        discovery_version,
        retries,
        interruptible,
        deprecated,
        storage_request,
        cpu_request,
        gpu_request,
        memory_request,
        storage_limit,
        cpu_limit,
        gpu_limit,
        memory_limit,
        discoverable,
        timeout,
        cluster_label,
        tags,
        environment,
    ):
        """
        :param task_function: Function container user code.  This will be executed via the SDK's engine.
        :param Text task_type: string describing the task type
        :param Text discovery_version: string describing the version for task discovery purposes
        :param int retries: Number of retries to attempt
        :param Text deprecated:
        :param Text storage_request:
        :param Text cpu_request:
        :param Text gpu_request:
        :param Text memory_request:
        :param Text storage_limit:
        :param Text cpu_limit:
        :param Text gpu_limit:
        :param Text memory_limit:
        :param bool discoverable:
        :param datetime.timedelta timeout:
        :param Text cluster_label:
        :param list[Text] tags:
        :param dict[Text, Text] environment:
        """
        self._task_function = task_function
        super(SdkHiveTask, self).__init__(
            task_function,
            task_type,
            discovery_version,
            retries,
            interruptible,
            deprecated,
            storage_request,
            cpu_request,
            gpu_request,
            memory_request,
            storage_limit,
            cpu_limit,
            gpu_limit,
            memory_limit,
            discoverable,
            timeout,
            environment,
            {},
        )
        self._validate_task_parameters(cluster_label, tags)
        self._cluster_label = cluster_label
        self._tags = tags

    def _generate_plugin_objects(self, context, inputs_dict):
        """
        Runs user code and and produces hive queries
        :param flytekit.engines.common.EngineContext context:
        :param dict[Text, T] inputs:
        :rtype: list[_qubole.QuboleHiveJob]
        """
        queries_from_task = super(SdkHiveTask, self)._execute_user_code(context, inputs_dict) or []
        if not isinstance(queries_from_task, list):
            queries_from_task = [queries_from_task]

        self._validate_queries(queries_from_task)
        plugin_objects = []

        for q in queries_from_task:
            hive_query = _qubole.HiveQuery(
                query=q,
                timeout_sec=self.metadata.timeout.seconds,
                retry_count=self.metadata.retries.retries,
            )

            # TODO: Remove this after all users of older SDK versions that did the single node, multi-query pattern are
            #       deprecated. This is only here for backwards compatibility - in addition to writing the query to the
            #       query field, we also construct a QueryCollection with only one query. This will ensure that the
            #       older plugin will continue to work.
            query_collection = _qubole.HiveQueryCollection([hive_query])

            plugin_objects.append(
                _qubole.QuboleHiveJob(
                    hive_query,
                    self._cluster_label,
                    self._tags,
                    query_collection=query_collection,
                )
            )

        return plugin_objects

    @staticmethod
    def _validate_task_parameters(cluster_label, tags):
        if not (cluster_label is None or isinstance(cluster_label, (str, _six.text_type))):
            raise _FlyteTypeException(
                type(cluster_label),
                {str, _six.text_type},
                additional_msg="cluster_label for a hive task must be in text format",
                received_value=cluster_label,
            )
        if tags is not None:
            if not (isinstance(tags, list) and all(isinstance(tag, (str, _six.text_type)) for tag in tags)):
                raise _FlyteTypeException(
                    type(tags),
                    [],
                    additional_msg="tags for a hive task must be in 'list of text' format",
                    received_value=tags,
                )
            if len(tags) > ALLOWED_TAGS_COUNT:
                raise _FlyteValueException(
                    len(tags),
                    "number of tags must be less than {}".format(ALLOWED_TAGS_COUNT),
                )
            if not all(len(tag) for tag in tags):
                raise _FlyteValueException(
                    tags,
                    "length of a tag must be less than {} chars".format(MAX_TAG_LENGTH),
                )

    @staticmethod
    def _validate_queries(queries_from_task):
        for query_from_task in queries_from_task or []:
            if not isinstance(query_from_task, (str, _six.text_type)):
                raise _FlyteTypeException(
                    type(query_from_task),
                    {str, _six.text_type},
                    additional_msg="All queries returned from a Hive task must be in text format.",
                    received_value=query_from_task,
                )

    def _produce_dynamic_job_spec(self, context, inputs):
        """
        Runs user code and and produces future task nodes to run sub-tasks.
        :param context:
        :param flytekit.models.literals.LiteralMap literal_map inputs:
        :rtype: flytekit.models.dynamic_job.DynamicJobSpec
        """
        inputs_dict = _type_helpers.unpack_literal_map_to_sdk_python_std(
            inputs,
            {k: _type_helpers.get_sdk_type_from_literal_type(v.type) for k, v in _six.iteritems(self.interface.inputs)},
        )
        outputs_dict = {
            name: _task_output.OutputReference(_type_helpers.get_sdk_type_from_literal_type(variable.type))
            for name, variable in _six.iteritems(self.interface.outputs)
        }

        # Add outputs to inputs
        inputs_dict.update(outputs_dict)

        nodes = []
        tasks = []
        # One node per query
        generated_queries = self._generate_plugin_objects(context, inputs_dict)

        # Create output bindings always - this has to happen after user code has run
        output_bindings = [
            _literal_models.Binding(
                var=name,
                binding=_interface.BindingData.from_python_std(b.sdk_type.to_flyte_literal_type(), b.value),
            )
            for name, b in _six.iteritems(outputs_dict)
        ]

        i = 0
        for quboleHiveJob in generated_queries:
            hive_job_node = _create_hive_job_node("HiveQuery_{}".format(i), quboleHiveJob.to_flyte_idl(), self.metadata)
            nodes.append(hive_job_node)
            tasks.append(hive_job_node.executable_sdk_object)
            i += 1

        dynamic_job_spec = _dynamic_job.DynamicJobSpec(
            min_successes=len(nodes),
            tasks=tasks,
            nodes=nodes,
            outputs=output_bindings,
            subworkflows=[],
        )

        return dynamic_job_spec

    @_exception_scopes.system_entry_point
    def execute(self, context, inputs):
        """
        Executes hive batch task's user code and produces futures file as well as all sub-task inputs.pb files.

        :param flytekit.engines.common.EngineContext context:
        :param flytekit.models.literals.LiteralMap inputs:
        :rtype: dict[Text, flytekit.models.common.FlyteIdlEntity]
        :returns: This function must return a dictionary mapping 'filenames' to Flyte Interface Entities.  These
            entities will be used by the engine to pass data from node to node, populate metadata, etc. etc..  Each
            engine will have different behavior.  For instance, the Flyte engine will upload the entities to a remote
            working directory (with the names provided), which will in turn allow Flyte Propeller to push along the
            workflow.  Where as local engine will merely feed the outputs directly into the next node.
        """
        spec = self._produce_dynamic_job_spec(context, inputs)
        generated_files = {}

        # If no queries were produced, then the spec should not have any nodes, in which case we just produce an
        # outputs file like any other single-step tasks.
        if len(spec.nodes) == 0:
            return {
                _constants.OUTPUT_FILE_NAME: _literal_models.LiteralMap(
                    literals={binding.var: binding.binding.to_literal_model() for binding in spec.outputs}
                )
            }
        else:
            generated_files.update({_constants.FUTURES_FILE_NAME: spec})

            return generated_files


def _create_hive_job_node(name, hive_job, metadata):
    """
    :param Text name:
    :param _qubole.QuboleHiveJob hive_job: Hive job spec
    :param flytekit.models.task.TaskMetadata metadata: This contains information needed at runtime to determine
        behavior such as whether or not outputs are discoverable, timeouts, and retries.
    :rtype: _nodes.SdkNode:
    """
    return _nodes.SdkNode(
        id=_six.text_type(_uuid.uuid4()),
        upstream_nodes=[],
        bindings=[],
        metadata=_workflow_model.NodeMetadata(name, metadata.timeout, _literal_models.RetryStrategy(0)),
        sdk_task=SdkHiveJob(hive_job, metadata),
    )


class SdkHiveJob(_base_task.SdkTask):
    """
    This class encapsulates the hive-job that is submitted to the Qubole Operator.

    """

    def __init__(
        self,
        hive_job,
        metadata,
    ):
        """
        :param _qubole.QuboleHiveJob hive_job: Hive job spec
        :param TaskMetadata metadata: This contains information needed at runtime to determine behavior such as
            whether or not outputs are discoverable, timeouts, and retries.
        """
        super(SdkHiveJob, self).__init__(
            _constants.SdkTaskType.HIVE_JOB,
            metadata,
            # Individual hive tasks never take anything, or return anything. They just run a query that's already
            # got the location set.
            _interface_model.TypedInterface({}, {}),
            _MessageToDict(hive_job),
        )



from flyteidl.admin import common_pb2 as _common

from flytekit.models import common as _common_models


class NamedEntityState(object):
    ACTIVE = _common.NAMED_ENTITY_ACTIVE
    ARCHIVED = _common.NAMED_ENTITY_ARCHIVED

    @classmethod
    def enum_to_string(cls, val):
        """
        :param int val:
        :rtype: Text
        """
        if val == cls.ACTIVE:
            return "ACTIVE"
        elif val == cls.ARCHIVED:
            return "ARCHIVED"
        else:
            return "<UNKNOWN>"


class NamedEntityIdentifier(_common_models.FlyteIdlEntity):
    def __init__(self, project, domain, name):
        """
        :param Text project:
        :param Text domain:
        :param Text name:
        """
        self._project = project
        self._domain = domain
        self._name = name

    @property
    def project(self):
        """
        :rtype: Text
        """
        return self._project

    @property
    def domain(self):
        """
        :rtype: Text
        """
        return self._domain

    @property
    def name(self):
        """
        :rtype: Text
        """
        return self._name

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.common_pb2.NamedEntityIdentifier
        """
        return _common.NamedEntityIdentifier(
            project=self.project,
            domain=self.domain,
            name=self.name,
        )

    @classmethod
    def from_flyte_idl(cls, p):
        """
        :param flyteidl.core.common_pb2.NamedEntityIdentifier p:
        :rtype: Identifier
        """
        return cls(
            project=p.project,
            domain=p.domain,
            name=p.name,
        )


class NamedEntityMetadata(_common_models.FlyteIdlEntity):
    def __init__(self, description, state):
        """

        :param Text description:
        :param int state: enum value from NamedEntityState
        """
        self._description = description
        self._state = state

    @property
    def description(self):
        """
        :rtype: Text
        """
        return self._description

    @property
    def state(self):
        """
        enum value from NamedEntityState
        :rtype: int
        """
        return self._state

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.common_pb2.NamedEntityMetadata
        """
        return _common.NamedEntityMetadata(
            description=self.description,
            state=self.state,
        )

    @classmethod
    def from_flyte_idl(cls, p):
        """
        :param flyteidl.core.common_pb2.NamedEntityMetadata p:
        :rtype: Identifier
        """
        return cls(
            description=p.description,
            state=p.state,
        )



import six as _six
from flyteidl.admin import common_pb2 as _common_pb2
from flyteidl.admin import execution_pb2 as _execution_pb2
from flyteidl.admin import launch_plan_pb2 as _launch_plan_pb2
from flyteidl.admin import matchable_resource_pb2 as _matchable_resource_pb2
from flyteidl.admin import node_execution_pb2 as _node_execution_pb2
from flyteidl.admin import project_domain_attributes_pb2 as _project_domain_attributes_pb2
from flyteidl.admin import project_pb2 as _project_pb2
from flyteidl.admin import task_execution_pb2 as _task_execution_pb2
from flyteidl.admin import task_pb2 as _task_pb2
from flyteidl.admin import workflow_attributes_pb2 as _workflow_attributes_pb2
from flyteidl.admin import workflow_pb2 as _workflow_pb2

from flytekit.clients.raw import RawSynchronousFlyteClient as _RawSynchronousFlyteClient
from flytekit.models import common as _common
from flytekit.models import execution as _execution
from flytekit.models import filters as _filters
from flytekit.models import launch_plan as _launch_plan
from flytekit.models import node_execution as _node_execution
from flytekit.models import project as _project
from flytekit.models import task as _task
from flytekit.models.admin import task_execution as _task_execution
from flytekit.models.admin import workflow as _workflow
from flytekit.models.core import identifier as _identifier


class SynchronousFlyteClient(_RawSynchronousFlyteClient):
    @property
    def raw(self):
        """
        Gives access to the raw client
        :rtype: flytekit.clients.raw.RawSynchronousFlyteClient
        """
        return super(SynchronousFlyteClient, self)

    ####################################################################################################################
    #
    #  Task Endpoints
    #
    ####################################################################################################################

    def create_task(self, task_identifer, task_spec):
        """
        This will create a task definition in the Admin database. Once successful, the task object can be
        retrieved via the client or viewed via the UI or command-line interfaces.

        .. note ::

            Overwrites are not supported so any request for a given project, domain, name, and version that exists in
            the database must match the existing definition exactly. Furthermore, as long as the request
            remains identical, calling this method multiple times will result in success.

        :param flytekit.models.core.identifier.Identifier task_identifer: The identifier for this task.
        :param flytekit.models.task.TaskSpec task_spec: This is the actual definition of the task that
            should be created.
        :raises flytekit.common.exceptions.user.FlyteEntityAlreadyExistsException: If an identical version of the
            task is found, this exception is raised.  The client might choose to ignore this exception because the
            identical task is already registered.
        :raises grpc.RpcError:
        """
        super(SynchronousFlyteClient, self).create_task(
            _task_pb2.TaskCreateRequest(id=task_identifer.to_flyte_idl(), spec=task_spec.to_flyte_idl())
        )

    def list_task_ids_paginated(self, project, domain, limit=100, token=None, sort_by=None):
        """
        This returns a page of identifiers for the tasks for a given project and domain. Filters can also be
        specified.

        .. note ::

            This is a paginated API.  Use the token field in the request to specify a page offset token.
            The user of the API is responsible for providing this token.

        .. note ::

            If entries are added to the database between requests for different pages, it is possible to receive
            entries on the second page that also appeared on the first.

        :param Text project: The namespace of the project to list.
        :param Text domain: The domain space of the project to list.
        :param int limit: [Optional] The maximum number of entries to return.  Must be greater than 0.  The maximum
            page size is determined by the Flyte Admin Service configuration.  If limit is greater than the maximum
            page size, an exception will be raised.
        :param Text token: [Optional] If specified, this specifies where in the rows of results to skip before reading.
            If you previously retrieved a page response with token="foo" and you want the next page,
            specify token="foo". Please see the notes for this function about the caveats of the paginated API.
        :param flytekit.models.admin.common.Sort sort_by: [Optional] If provided, the results will be sorted.
        :raises: TODO
        :rtype: list[flytekit.models.common.NamedEntityIdentifier], Text
        """
        identifier_list = super(SynchronousFlyteClient, self).list_task_ids_paginated(
            _common_pb2.NamedEntityIdentifierListRequest(
                project=project,
                domain=domain,
                limit=limit,
                token=token,
                sort_by=None if sort_by is None else sort_by.to_flyte_idl(),
            )
        )
        return (
            [_common.NamedEntityIdentifier.from_flyte_idl(identifier_pb) for identifier_pb in identifier_list.entities],
            _six.text_type(identifier_list.token),
        )

    def list_tasks_paginated(self, identifier, limit=100, token=None, filters=None, sort_by=None):
        """
        This returns a page of task metadata for tasks in a given project and domain.  Optionally,
        specifying a name will limit the results to only tasks with that name in the given project and domain.

        .. note ::

            This is a paginated API.  Use the token field in the request to specify a page offset token.
            The user of the API is responsible for providing this token.

        .. note ::

            If entries are added to the database between requests for different pages, it is possible to receive
            entries on the second page that also appeared on the first.

        :param flytekit.models.common.NamedEntityIdentifier identifier: NamedEntityIdentifier to list.
        :param int limit: [Optional] The maximum number of entries to return.  Must be greater than 0.  The maximum
            page size is determined by the Flyte Admin Service configuration.  If limit is greater than the maximum
            page size, an exception will be raised.
        :param int token: [Optional] If specified, this specifies where in the rows of results to skip before reading.
            If you previously retrieved a page response with token="foo" and you want the next page,
            specify token="foo". Please see the notes for this function about the caveats of the paginated API.
        :param list[flytekit.models.filters.Filter] filters: [Optional] If specified, the filters will be applied to
            the query.  If the filter is not supported, an exception will be raised.
        :param flytekit.models.admin.common.Sort sort_by: [Optional] If provided, the results will be sorted.
        :raises: TODO
        :rtype: list[flytekit.models.task.Task], Text
        """
        task_list = super(SynchronousFlyteClient, self).list_tasks_paginated(
            resource_list_request=_common_pb2.ResourceListRequest(
                id=identifier.to_flyte_idl(),
                limit=limit,
                token=token,
                filters=_filters.FilterList(filters or []).to_flyte_idl(),
                sort_by=None if sort_by is None else sort_by.to_flyte_idl(),
            )
        )
        # TODO: tmp workaround
        for pb in task_list.tasks:
            pb.id.resource_type = _identifier.ResourceType.TASK
        return (
            [_task.Task.from_flyte_idl(task_pb2) for task_pb2 in task_list.tasks],
            _six.text_type(task_list.token),
        )

    def get_task(self, id):
        """
        This returns a single task for a given identifier.

        :param flytekit.models.core.identifier.Identifier id: The ID representing a given task.
        :raises: TODO
        :rtype: flytekit.models.task.Task
        """
        return _task.Task.from_flyte_idl(
            super(SynchronousFlyteClient, self).get_task(_common_pb2.ObjectGetRequest(id=id.to_flyte_idl()))
        )

    ####################################################################################################################
    #
    #  Workflow Endpoints
    #
    ####################################################################################################################

    def create_workflow(self, workflow_identifier, workflow_spec):
        """
        This will create a workflow definition in the Admin database. Once successful, the workflow object can be
        retrieved via the client or viewed via the UI or command-line interfaces.

        .. note ::

            Overwrites are not supported so any request for a given project, domain, name, and version that exists in
            the database must match the existing definition exactly. Furthermore, as long as the request
            remains identical, calling this method multiple times will result in success.

        :param: flytekit.models.core.identifier.Identifier workflow_identifier: The identifier for this workflow.
        :param: flytekit.models.admin.workflow.WorkflowSpec workflow_spec: This is the actual definition of the workflow
            that should be created.
        :raises flytekit.common.exceptions.user.FlyteEntityAlreadyExistsException: If an identical version of the
            workflow is found, this exception is raised.  The client might choose to ignore this exception because the
            identical workflow is already registered.
        :raises grpc.RpcError:
        """
        super(SynchronousFlyteClient, self).create_workflow(
            _workflow_pb2.WorkflowCreateRequest(
                id=workflow_identifier.to_flyte_idl(), spec=workflow_spec.to_flyte_idl()
            )
        )

    def list_workflow_ids_paginated(self, project, domain, limit=100, token=None, sort_by=None):
        """
        This returns a page of identifiers for the workflows for a given project and domain. Filters can also be
        specified.

        .. note ::

            This is a paginated API.  Use the token field in the request to specify a page offset token.
            The user of the API is responsible for providing this token.

        .. note ::

            If entries are added to the database between requests for different pages, it is possible to receive
            entries on the second page that also appeared on the first.

        :param: Text project: The namespace of the project to list.
        :param: Text domain: The domain space of the project to list.
        :param: int limit: [Optional] The maximum number of entries to return.  Must be greater than 0.  The maximum
            page size is determined by the Flyte Admin Service configuration.  If limit is greater than the maximum
            page size, an exception will be raised.
        :param: int token: [Optional] If specified, this specifies where in the rows of results to skip before reading.
            If you previously retrieved a page response with token="foo" and you want the next page,
            specify token="foo". Please see the notes for this function about the caveats of the paginated API.
        :param flytekit.models.admin.common.Sort sort_by: [Optional] If provided, the results will be sorted.
        :raises: TODO
        :rtype: list[flytekit.models.common.NamedEntityIdentifier], Text
        """
        identifier_list = super(SynchronousFlyteClient, self).list_workflow_ids_paginated(
            _common_pb2.NamedEntityIdentifierListRequest(
                project=project,
                domain=domain,
                limit=limit,
                token=token,
                sort_by=None if sort_by is None else sort_by.to_flyte_idl(),
            )
        )
        return (
            [_common.NamedEntityIdentifier.from_flyte_idl(identifier_pb) for identifier_pb in identifier_list.entities],
            _six.text_type(identifier_list.token),
        )

    def list_workflows_paginated(self, identifier, limit=100, token=None, filters=None, sort_by=None):
        """
        This returns a page of workflow meta-information for workflows in a given project and domain.  Optionally,
        specifying a name will limit the results to only workflows with that name in the given project and domain.

        .. note ::

            This is a paginated API.  Use the token field in the request to specify a page offset token.
            The user of the API is responsible for providing this token.

        .. note ::

            If entries are added to the database between requests for different pages, it is possible to receive
            entries on the second page that also appeared on the first.

        :param flytekit.models.common.NamedEntityIdentifier identifier: NamedEntityIdentifier to list.
        :param int limit: [Optional] The maximum number of entries to return.  Must be greater than 0.  The maximum
            page size is determined by the Flyte Admin Service configuration.  If limit is greater than the maximum
            page size, an exception will be raised.
        :param int token: [Optional] If specified, this specifies where in the rows of results to skip before reading.
            If you previously retrieved a page response with token="foo" and you want the next page,
            specify token="foo". Please see the notes for this function about the caveats of the paginated API.
        :param list[flytekit.models.filters.Filter] filters: [Optional] If specified, the filters will be applied to
            the query.  If the filter is not supported, an exception will be raised.
        :param flytekit.models.admin.common.Sort sort_by: [Optional] If provided, the results will be sorted.
        :raises: TODO
        :rtype: list[flytekit.models.admin.workflow.Workflow], Text
        """
        wf_list = super(SynchronousFlyteClient, self).list_workflows_paginated(
            resource_list_request=_common_pb2.ResourceListRequest(
                id=identifier.to_flyte_idl(),
                limit=limit,
                token=token,
                filters=_filters.FilterList(filters or []).to_flyte_idl(),
                sort_by=None if sort_by is None else sort_by.to_flyte_idl(),
            )
        )
        # TODO: tmp workaround
        for pb in wf_list.workflows:
            pb.id.resource_type = _identifier.ResourceType.WORKFLOW
        return (
            [_workflow.Workflow.from_flyte_idl(wf_pb2) for wf_pb2 in wf_list.workflows],
            _six.text_type(wf_list.token),
        )

    def get_workflow(self, id):
        """
        This returns a single task for a given ID.

        :param flytekit.models.core.identifier.Identifier id: The ID representing a given task.
        :raises: TODO
        :rtype: flytekit.models.admin.workflow.Workflow
        """
        return _workflow.Workflow.from_flyte_idl(
            super(SynchronousFlyteClient, self).get_workflow(_common_pb2.ObjectGetRequest(id=id.to_flyte_idl()))
        )

    ####################################################################################################################
    #
    #  Launch Plan Endpoints
    #
    ####################################################################################################################

    def create_launch_plan(self, launch_plan_identifer, launch_plan_spec):
        """
        This will create a launch plan definition in the Admin database.  Once successful, the launch plan object can be
        retrieved via the client or viewed via the UI or command-line interfaces.

        .. note ::

            Overwrites are not supported so any request for a given project, domain, name, and version that exists in
            the database must match the existing definition exactly.  This also means that as long as the request
            remains identical, calling this method multiple times will result in success.

        :param: flytekit.models.core.identifier.Identifier launch_plan_identifer: The identifier for this launch plan.
        :param: flytekit.models.launch_plan.LaunchPlanSpec launch_plan_spec: This is the actual definition of the
            launch plan that should be created.
        :raises flytekit.common.exceptions.user.FlyteEntityAlreadyExistsException: If an identical version of the
            launch plan is found, this exception is raised.  The client might choose to ignore this exception because
            the identical launch plan is already registered.
        :raises grpc.RpcError:
        """
        super(SynchronousFlyteClient, self).create_launch_plan(
            _launch_plan_pb2.LaunchPlanCreateRequest(
                id=launch_plan_identifer.to_flyte_idl(),
                spec=launch_plan_spec.to_flyte_idl(),
            )
        )

    def get_launch_plan(self, id):
        """
        Retrieves a launch plan entity.

        :param flytekit.models.core.identifier.Identifier id: unique identifier for launch plan to retrieve
        :rtype: flytekit.models.launch_plan.LaunchPlan
        """
        return _launch_plan.LaunchPlan.from_flyte_idl(
            super(SynchronousFlyteClient, self).get_launch_plan(_common_pb2.ObjectGetRequest(id=id.to_flyte_idl()))
        )

    def get_active_launch_plan(self, identifier):
        """
        Retrieves the active launch plan entity given a named entity identifier (project, domain, name).  Raises an
        error if no active launch plan exists.

        :param flytekit.models.common.NamedEntityIdentifier identifier: NamedEntityIdentifier to list.
        :rtype: flytekit.models.launch_plan.LaunchPlan
        """
        return _launch_plan.LaunchPlan.from_flyte_idl(
            super(SynchronousFlyteClient, self).get_active_launch_plan(
                _launch_plan_pb2.ActiveLaunchPlanRequest(id=identifier.to_flyte_idl())
            )
        )

    def list_launch_plan_ids_paginated(self, project, domain, limit=100, token=None, sort_by=None):
        """
        This returns a page of identifiers for the launch plans for a given project and domain. Filters can also be
        specified.

        .. note ::

            This is a paginated API.  Use the token field in the request to specify a page offset token.
            The user of the API is responsible for providing this token.

        .. note ::

            If entries are added to the database between requests for different pages, it is possible to receive
            entries on the second page that also appeared on the first.

        :param: Text project: The namespace of the project to list.
        :param: Text domain: The domain space of the project to list.
        :param: int limit: [Optional] The maximum number of entries to return.  Must be greater than 0.  The maximum
            page size is determined by the Flyte Admin Service configuration.  If limit is greater than the maximum
            page size, an exception will be raised.
        :param: int token: [Optional] If specified, this specifies where in the rows of results to skip before reading.
            If you previously retrieved a page response with token="foo" and you want the next page,
            specify token="foo". Please see the notes for this function about the caveats of the paginated API.
        :param flytekit.models.admin.common.Sort sort_by: [Optional] If provided, the results will be sorted.
        :raises: TODO
        :rtype: list[flytekit.models.common.NamedEntityIdentifier], Text
        """
        identifier_list = super(SynchronousFlyteClient, self).list_launch_plan_ids_paginated(
            _common_pb2.NamedEntityIdentifierListRequest(
                project=project,
                domain=domain,
                limit=limit,
                token=token,
                sort_by=None if sort_by is None else sort_by.to_flyte_idl(),
            )
        )
        return (
            [_common.NamedEntityIdentifier.from_flyte_idl(identifier_pb) for identifier_pb in identifier_list.entities],
            _six.text_type(identifier_list.token),
        )

    def list_launch_plans_paginated(self, identifier, limit=100, token=None, filters=None, sort_by=None):
        """
        This returns a page of launch plan meta-information for launch plans in a given project and domain.  Optionally,
        specifying a name will limit the results to only workflows with that name in the given project and domain.

        .. note ::

            This is a paginated API.  Use the token field in the request to specify a page offset token.
            The user of the API is responsible for providing this token.

        .. note ::

            If entries are added to the database between requests for different pages, it is possible to receive
            entries on the second page that also appeared on the first.

        :param flytekit.models.common.NamedEntityIdentifier identifier: NamedEntityIdentifier to list.
        :param int limit: [Optional] The maximum number of entries to return.  Must be greater than 0.  The maximum
            page size is determined by the Flyte Admin Service configuration.  If limit is greater than the maximum
            page size, an exception will be raised.
        :param int token: [Optional] If specified, this specifies where in the rows of results to skip before reading.
            If you previously retrieved a page response with token="foo" and you want the next page,
            specify token="foo". Please see the notes for this function about the caveats of the paginated API.
        :param list[flytekit.models.filters.Filter] filters: [Optional] If specified, the filters will be applied to
            the query.  If the filter is not supported, an exception will be raised.
        :param flytekit.models.admin.common.Sort sort_by: [Optional] If provided, the results will be sorted.
        :raises: TODO
        :rtype: list[flytekit.models.launch_plan.LaunchPlan], str
        """
        lp_list = super(SynchronousFlyteClient, self).list_launch_plans_paginated(
            resource_list_request=_common_pb2.ResourceListRequest(
                id=identifier.to_flyte_idl(),
                limit=limit,
                token=token,
                filters=_filters.FilterList(filters or []).to_flyte_idl(),
                sort_by=None if sort_by is None else sort_by.to_flyte_idl(),
            )
        )
        # TODO: tmp workaround
        for pb in lp_list.launch_plans:
            pb.id.resource_type = _identifier.ResourceType.LAUNCH_PLAN
        return (
            [_launch_plan.LaunchPlan.from_flyte_idl(pb) for pb in lp_list.launch_plans],
            _six.text_type(lp_list.token),
        )

    def list_active_launch_plans_paginated(self, project, domain, limit=100, token=None, sort_by=None):
        """
        This returns a page of currently active launch plan meta-information for launch plans in a given project and
        domain.

        .. note ::

            This is a paginated API.  Use the token field in the request to specify a page offset token.
            The user of the API is responsible for providing this token.

        .. note ::

            If entries are added to the database between requests for different pages, it is possible to receive
            entries on the second page that also appeared on the first.

        :param Text project:
        :param Text domain:
        :param int limit: [Optional] The maximum number of entries to return.  Must be greater than 0.  The maximum
            page size is determined by the Flyte Admin Service configuration.  If limit is greater than the maximum
            page size, an exception will be raised.
        :param int token: [Optional] If specified, this specifies where in the rows of results to skip before reading.
            If you previously retrieved a page response with token="foo" and you want the next page,
            specify token="foo". Please see the notes for this function about the caveats of the paginated API.
        :param flytekit.models.admin.common.Sort sort_by: [Optional] If provided, the results will be sorted.
        :raises: TODO
        :rtype: list[flytekit.models.launch_plan.LaunchPlan], str
        """
        lp_list = super(SynchronousFlyteClient, self).list_active_launch_plans_paginated(
            _launch_plan_pb2.ActiveLaunchPlanListRequest(
                project=project,
                domain=domain,
                limit=limit,
                token=token,
                sort_by=None if sort_by is None else sort_by.to_flyte_idl(),
            )
        )
        # TODO: tmp workaround
        for pb in lp_list.launch_plans:
            pb.id.resource_type = _identifier.ResourceType.LAUNCH_PLAN
        return (
            [_launch_plan.LaunchPlan.from_flyte_idl(pb) for pb in lp_list.launch_plans],
            _six.text_type(lp_list.token),
        )

    def update_launch_plan(self, id, state):
        """
        Updates a launch plan.  Currently, this can only be used to update a given launch plan's state (ACTIVE v.
        INACTIVE) for schedules.  If a launch plan with a given project, domain, and name is set to ACTIVE,
        then any other launch plan with the same project, domain, and name that was set to ACTIVE will be switched to
        INACTIVE in one transaction.

        :param flytekit.models.core.identifier.Identifier id: identifier for launch plan to update
        :param int state: Enum value from flytekit.models.launch_plan.LaunchPlanState
        """
        super(SynchronousFlyteClient, self).update_launch_plan(
            _launch_plan_pb2.LaunchPlanUpdateRequest(id=id.to_flyte_idl(), state=state)
        )

    ####################################################################################################################
    #
    #  Named Entity Endpoints
    #
    ####################################################################################################################

    def update_named_entity(self, resource_type, id, metadata):
        """
        Updates the metadata associated with a named entity.  A named entity is designated a resource, e.g. a workflow,
        task or launch plan specified by {project, domain, name} across all versions of the resource.

        :param int resource_type: Enum value from flytekit.models.identifier.ResourceType
        :param flytekit.models.admin.named_entity.NamedEntityIdentifier id: identifier for named entity to update
        :param flytekit.models.admin.named_entity.NamedEntityIdentifierMetadata metadata:
        """
        super(SynchronousFlyteClient, self).update_named_entity(
            _common_pb2.NamedEntityUpdateRequest(
                resource_type=resource_type,
                id=id.to_flyte_idl(),
                metadata=metadata.to_flyte_idl(),
            )
        )

    ####################################################################################################################
    #
    #  Execution Endpoints
    #
    ####################################################################################################################

    def create_execution(self, project, domain, name, execution_spec, inputs):
        """
        This will create an execution for the given execution spec.
        :param Text project:
        :param Text domain:
        :param Text name:
        :param flytekit.models.execution.ExecutionSpec execution_spec: This is the specification for the execution.
        :param flytekit.models.literals.LiteralMap inputs: The inputs for the execution
        :returns: The unique identifier for the execution.
        :rtype: flytekit.models.core.identifier.WorkflowExecutionIdentifier
        """
        return _identifier.WorkflowExecutionIdentifier.from_flyte_idl(
            super(SynchronousFlyteClient, self)
            .create_execution(
                _execution_pb2.ExecutionCreateRequest(
                    project=project,
                    domain=domain,
                    name=name,
                    spec=execution_spec.to_flyte_idl(),
                    inputs=inputs.to_flyte_idl(),
                )
            )
            .id
        )

    def get_execution(self, id):
        """
        :param flytekit.common.core.identifier.WorkflowExecutionIdentifier id:
        :rtype: flytekit.models.execution.Execution
        """
        return _execution.Execution.from_flyte_idl(
            super(SynchronousFlyteClient, self).get_execution(
                _execution_pb2.WorkflowExecutionGetRequest(id=id.to_flyte_idl())
            )
        )

    def get_execution_data(self, id):
        """
        Returns signed URLs to LiteralMap blobs for an execution's inputs and outputs (when available).

        :param flytekit.models.core.identifier.WorkflowExecutionIdentifier id:
        :rtype: flytekit.models.execution.WorkflowExecutionGetDataResponse
        """
        return _execution.WorkflowExecutionGetDataResponse.from_flyte_idl(
            super(SynchronousFlyteClient, self).get_execution_data(
                _execution_pb2.WorkflowExecutionGetDataRequest(id=id.to_flyte_idl())
            )
        )

    def list_executions_paginated(self, project, domain, limit=100, token=None, filters=None, sort_by=None):
        """
        This returns a page of executions in a given project and domain.

        .. note ::

            This is a paginated API.  Use the token field in the request to specify a page offset token.
            The user of the API is responsible for providing this token.

        .. note ::

            If entries are added to the database between requests for different pages, it is possible to receive
            entries on the second page that also appeared on the first.

        :param Text project: Project in which to list executions.
        :param Text domain: Project in which to list executions.
        :param int limit: [Optional] The maximum number of entries to return.  Must be greater than 0.  The maximum
            page size is determined by the Flyte Admin Service configuration.  If limit is greater than the maximum
            page size, an exception will be raised.
        :param Text token: [Optional] If specified, this specifies where in the rows of results to skip before reading.
            If you previously retrieved a page response with token="foo" and you want the next page,
            specify token="foo". Please see the notes for this function about the caveats of the paginated API.
        :param list[flytekit.models.filters.Filter] filters: [Optional] If specified, the filters will be applied to
            the query.  If the filter is not supported, an exception will be raised.
        :param flytekit.models.admin.common.Sort sort_by: [Optional] If provided, the results will be sorted.
        :raises: TODO
        :rtype: (list[flytekit.models.execution.Execution], Text)
        """
        exec_list = super(SynchronousFlyteClient, self).list_executions_paginated(
            resource_list_request=_common_pb2.ResourceListRequest(
                id=_common_pb2.NamedEntityIdentifier(project=project, domain=domain),
                limit=limit,
                token=token,
                filters=_filters.FilterList(filters or []).to_flyte_idl(),
                sort_by=None if sort_by is None else sort_by.to_flyte_idl(),
            )
        )
        return (
            [_execution.Execution.from_flyte_idl(pb) for pb in exec_list.executions],
            _six.text_type(exec_list.token),
        )

    def terminate_execution(self, id, cause):
        """
        :param flytekit.common.core.identifier.WorkflowExecutionIdentifier id:
        :param Text cause:
        """
        super(SynchronousFlyteClient, self).terminate_execution(
            _execution_pb2.ExecutionTerminateRequest(id=id.to_flyte_idl(), cause=cause)
        )

    def relaunch_execution(self, id, name=None):
        """
        :param flytekit.common.core.identifier.WorkflowExecutionIdentifier id:
        :param Text name: [Optional] name for the new execution. If not specified, a randomly generated name will be
            used
        :returns: The unique identifier for the new execution.
        :rtype: flytekit.models.core.identifier.WorkflowExecutionIdentifier
        """
        return _identifier.WorkflowExecutionIdentifier.from_flyte_idl(
            super(SynchronousFlyteClient, self)
            .relaunch_execution(_execution_pb2.ExecutionRelaunchRequest(id=id.to_flyte_idl(), name=name))
            .id
        )

    ####################################################################################################################
    #
    #  Node Execution Endpoints
    #
    ####################################################################################################################

    def get_node_execution(self, node_execution_identifier):
        """
        :param flytekit.models.core.identifier.NodeExecutionIdentifier node_execution_identifier:
        :rtype: flytekit.models.node_execution.NodeExecution
        """
        return _node_execution.NodeExecution.from_flyte_idl(
            super(SynchronousFlyteClient, self).get_node_execution(
                _node_execution_pb2.NodeExecutionGetRequest(id=node_execution_identifier.to_flyte_idl())
            )
        )

    def get_node_execution_data(self, node_execution_identifier):
        """
        Returns signed URLs to LiteralMap blobs for a node execution's inputs and outputs (when available).

        :param flytekit.models.core.identifier.NodeExecutionIdentifier node_execution_identifier:
        :rtype: flytekit.models.execution.NodeExecutionGetDataResponse
        """
        return _execution.NodeExecutionGetDataResponse.from_flyte_idl(
            super(SynchronousFlyteClient, self).get_node_execution_data(
                _node_execution_pb2.NodeExecutionGetDataRequest(id=node_execution_identifier.to_flyte_idl())
            )
        )

    def list_node_executions(
        self,
        workflow_execution_identifier,
        limit=100,
        token=None,
        filters=None,
        sort_by=None,
    ):
        """
        TODO: Comment
        :param flytekit.models.core.identifier.WorkflowExecutionIdentifier workflow_execution_identifier:
        :param int limit:
        :param Text token: [Optional] If specified, this specifies where in the rows of results to skip before reading.
        If you previously retrieved a page response with token="foo" and you want the next page,
        specify token="foo".
        :param list[flytekit.models.filters.Filter] filters:
        :param flytekit.models.admin.common.Sort sort_by: [Optional] If provided, the results will be sorted.
        :rtype: list[flytekit.models.node_execution.NodeExecution], Text
        """
        exec_list = super(SynchronousFlyteClient, self).list_node_executions_paginated(
            _node_execution_pb2.NodeExecutionListRequest(
                workflow_execution_id=workflow_execution_identifier.to_flyte_idl(),
                limit=limit,
                token=token,
                filters=_filters.FilterList(filters or []).to_flyte_idl(),
                sort_by=None if sort_by is None else sort_by.to_flyte_idl(),
            )
        )
        return (
            [_node_execution.NodeExecution.from_flyte_idl(e) for e in exec_list.node_executions],
            _six.text_type(exec_list.token),
        )

    def list_node_executions_for_task_paginated(
        self,
        task_execution_identifier,
        limit=100,
        token=None,
        filters=None,
        sort_by=None,
    ):
        """
        This returns nodes spawned by a specific task execution.  This is generally from things like dynamic tasks.
        :param flytekit.models.core.identifier.TaskExecutionIdentifier task_execution_identifier:
        :param int limit: Number to return per page
        :param Text token: [Optional] If specified, this specifies where in the rows of results to skip before reading.
        If you previously retrieved a page response with token="foo" and you want the next page,
        specify token="foo".
        :param list[flytekit.models.filters.Filter] filters:
        :param flytekit.models.admin.common.Sort sort_by: [Optional] If provided, the results will be sorted.
        :rtype: list[flytekit.models.node_execution.NodeExecution], Text
        """
        exec_list = self._stub.ListNodeExecutionsForTask(
            _node_execution_pb2.NodeExecutionForTaskListRequest(
                task_execution_id=task_execution_identifier.to_flyte_idl(),
                limit=limit,
                token=token,
                filters=_filters.FilterList(filters or []).to_flyte_idl(),
                sort_by=None if sort_by is None else sort_by.to_flyte_idl(),
            )
        )
        return (
            [_node_execution.NodeExecution.from_flyte_idl(e) for e in exec_list.node_executions],
            _six.text_type(exec_list.token),
        )

    ####################################################################################################################
    #
    #  Task Execution Endpoints
    #
    ####################################################################################################################

    def get_task_execution(self, id):
        """
        :param flytekit.models.core.identifier.TaskExecutionIdentifier id:
        :rtype: flytekit.models.admin.task_execution.TaskExecution
        """
        return _task_execution.TaskExecution.from_flyte_idl(
            super(SynchronousFlyteClient, self).get_task_execution(
                _task_execution_pb2.TaskExecutionGetRequest(id=id.to_flyte_idl())
            )
        )

    def get_task_execution_data(self, task_execution_identifier):
        """
        Returns signed URLs to LiteralMap blobs for a node execution's inputs and outputs (when available).

        :param flytekit.models.core.identifier.TaskExecutionIdentifier task_execution_identifier:
        :rtype: flytekit.models.execution.NodeExecutionGetDataResponse
        """
        return _execution.TaskExecutionGetDataResponse.from_flyte_idl(
            super(SynchronousFlyteClient, self).get_task_execution_data(
                _task_execution_pb2.TaskExecutionGetDataRequest(id=task_execution_identifier.to_flyte_idl())
            )
        )

    def list_task_executions_paginated(
        self,
        node_execution_identifier,
        limit=100,
        token=None,
        filters=None,
        sort_by=None,
    ):
        """
        :param flytekit.models.core.identifier.NodeExecutionIdentifier node_execution_identifier:
        :param int limit:
        :param Text token: [Optional] If specified, this specifies where in the rows of results to skip before reading.
        If you previously retrieved a page response with token="foo" and you want the next page,
        specify token="foo".
        :param list[flytekit.models.filters.Filter] filters:
        :param flytekit.models.admin.common.Sort sort_by: [Optional] If provided, the results will be sorted.
        :rtype: (list[flytekit.models.admin.task_execution.TaskExecution], Text)
        """
        exec_list = super(SynchronousFlyteClient, self).list_task_executions_paginated(
            _task_execution_pb2.TaskExecutionListRequest(
                node_execution_id=node_execution_identifier.to_flyte_idl(),
                limit=limit,
                token=token,
                filters=_filters.FilterList(filters or []).to_flyte_idl(),
                sort_by=None if sort_by is None else sort_by.to_flyte_idl(),
            )
        )
        return (
            [_task_execution.TaskExecution.from_flyte_idl(e) for e in exec_list.task_executions],
            _six.text_type(exec_list.token),
        )

    ####################################################################################################################
    #
    #  Project Endpoints
    #
    ####################################################################################################################

    def register_project(self, project):
        """
        Registers a project.
        :param flytekit.models.project.Project project:
        :rtype: flyteidl.admin.project_pb2.ProjectRegisterResponse
        """
        super(SynchronousFlyteClient, self).register_project(
            _project_pb2.ProjectRegisterRequest(
                project=project.to_flyte_idl(),
            )
        )

    def update_project(self, project):
        """
        Update an existing project specified by id.
        :param flytekit.models.project.Project project:
        :rtype: flyteidl.admin.project_pb2.ProjectUpdateResponse
        """
        super(SynchronousFlyteClient, self).update_project(project.to_flyte_idl())

    def list_projects_paginated(self, limit=100, token=None, filters=None, sort_by=None):
        """
        This returns a page of projects.

        .. note ::

            This is a paginated API.  Use the token field in the request to specify a page offset token.
            The user of the API is responsible for providing this token.

        .. note ::

            If entries are added to the database between requests for different pages, it is possible to receive
            entries on the second page that also appeared on the first.

        :param int limit: [Optional] The maximum number of entries to return.  Must be greater than 0.  The maximum
            page size is determined by the Flyte Admin Service configuration.  If limit is greater than the maximum
            page size, an exception will be raised.
        :param Text token: [Optional] If specified, this specifies where in the rows of results to skip before reading.
            If you previously retrieved a page response with token="foo" and you want the next page,
            specify token="foo". Please see the notes for this function about the caveats of the paginated API.
        :param list[flytekit.models.filters.Filter] filters: [Optional] If specified, the filters will be applied to
            the query.  If the filter is not supported, an exception will be raised.
        :param flytekit.models.admin.common.Sort sort_by: [Optional] If provided, the results will be sorted.
        :raises grpc.RpcError:
        :rtype: (list[flytekit.models.Project], Text)
        """
        projects = super(SynchronousFlyteClient, self).list_projects(
            project_list_request=_project_pb2.ProjectListRequest(
                limit=limit,
                token=token,
                filters=_filters.FilterList(filters or []).to_flyte_idl(),
                sort_by=None if sort_by is None else sort_by.to_flyte_idl(),
            )
        )
        return (
            [_project.Project.from_flyte_idl(pb) for pb in projects.projects],
            _six.text_type(projects.token),
        )

    ####################################################################################################################
    #
    #  Matching Attributes Endpoints
    #
    ####################################################################################################################

    def update_project_domain_attributes(self, project, domain, matching_attributes):
        """
        Sets custom attributes for a project and domain combination.
        :param Text project:
        :param Text domain:
        :param flytekit.models.MatchingAttributes matching_attributes:
        :return:
        """
        super(SynchronousFlyteClient, self).update_project_domain_attributes(
            _project_domain_attributes_pb2.ProjectDomainAttributesUpdateRequest(
                attributes=_project_domain_attributes_pb2.ProjectDomainAttributes(
                    project=project,
                    domain=domain,
                    matching_attributes=matching_attributes.to_flyte_idl(),
                )
            )
        )

    def update_workflow_attributes(self, project, domain, workflow, matching_attributes):
        """
        Sets custom attributes for a project, domain, and workflow combination.
        :param Text project:
        :param Text domain:
        :param Text workflow:
        :param flytekit.models.MatchingAttributes matching_attributes:
        :return:
        """
        super(SynchronousFlyteClient, self).update_workflow_attributes(
            _workflow_attributes_pb2.WorkflowAttributesUpdateRequest(
                attributes=_workflow_attributes_pb2.WorkflowAttributes(
                    project=project,
                    domain=domain,
                    workflow=workflow,
                    matching_attributes=matching_attributes.to_flyte_idl(),
                )
            )
        )

    def get_project_domain_attributes(self, project, domain, resource_type):
        """
        Fetches the custom attributes set for a project and domain combination.
        :param Text project:
        :param Text domain:
        :param flytekit.models.MatchableResource resource_type:
        :return:
        """
        return super(SynchronousFlyteClient, self).get_project_domain_attributes(
            _project_domain_attributes_pb2.ProjectDomainAttributesGetRequest(
                project=project,
                domain=domain,
                resource_type=resource_type,
            )
        )

    def get_workflow_attributes(self, project, domain, workflow, resource_type):
        """
        Fetches the custom attributes set for a project, domain, and workflow combination.
        :param Text project:
        :param Text domain:
        :param Text workflow:
        :param flytekit.models.MatchableResource resource_type:
        :return:
        """
        return super(SynchronousFlyteClient, self).get_workflow_attributes(
            _workflow_attributes_pb2.WorkflowAttributesGetRequest(
                project=project,
                domain=domain,
                workflow=workflow,
                resource_type=resource_type,
            )
        )

    def list_matchable_attributes(self, resource_type):
        """
        Fetches all custom attributes for a resource type.
        :param flytekit.models.MatchableResource resource_type:
        :return:
        """
        return super(SynchronousFlyteClient, self).list_matchable_attributes(
            _matchable_resource_pb2.ListMatchableAttributesRequest(
                resource_type=resource_type,
            )
        )



import importlib as _importlib
import inspect
import inspect as _inspect
import logging as _logging
from typing import Callable

from flytekit.common.exceptions import system as _system_exceptions


class InstanceTrackingMeta(type):
    """
    Please see the original class :py:class`flytekit.common.mixins.registerable._InstanceTracker` also and also look
    at the tests in the ``tests/flytekit/unit/core/tracker/test_tracking/`` folder to see how it's used.

    Basically, this will make instances of classes that use this metaclass aware of the module (the .py file) that
    caused the instance to be created. This is useful because it means that we can then (at least try to) find the
    variable that the instance was assigned to.
    """

    @staticmethod
    def _find_instance_module():
        frame = _inspect.currentframe()
        while frame:
            if frame.f_code.co_name == "<module>":
                return frame.f_globals["__name__"]
            frame = frame.f_back
        return None

    def __call__(cls, *args, **kwargs):
        o = super(InstanceTrackingMeta, cls).__call__(*args, **kwargs)
        o._instantiated_in = InstanceTrackingMeta._find_instance_module()
        return o


class TrackedInstance(metaclass=InstanceTrackingMeta):
    """
    Please see the notes for the metaclass above first.

    This functionality has two use-cases currently,
    * Keep track of naming for non-function ``PythonAutoContainerTasks``.  That is, things like the
      :py:class:`flytekit.extras.sqlite3.task.SQLite3Task` task.
    * Task resolvers, because task resolvers are instances of :py:class:`flytekit.core.python_auto_container.TaskResolverMixin`
      classes, not the classes themselves, which means we need to look on the left hand side of them to see how to
      find them at task execution time.
    """

    def __init__(self, *args, **kwargs):
        self._instantiated_in = None
        self._lhs = None
        super().__init__(*args, **kwargs)

    @property
    def instantiated_in(self) -> str:
        return self._instantiated_in

    @property
    def location(self) -> str:
        return f"{self.instantiated_in}.{self.lhs}"

    @property
    def lhs(self):
        if self._lhs is not None:
            return self._lhs
        return self.find_lhs()

    def find_lhs(self) -> str:
        if self._lhs is not None:
            return self._lhs

        if self._instantiated_in is None or self._instantiated_in == "":
            raise _system_exceptions.FlyteSystemException(f"Object {self} does not have an _instantiated in")

        _logging.debug(f"Looking for LHS for {self} from {self._instantiated_in}")
        m = _importlib.import_module(self._instantiated_in)
        for k in dir(m):
            try:
                if getattr(m, k) is self:
                    _logging.debug(f"Found LHS for {self}, {k}")
                    self._lhs = k
                    return k
            except ValueError as err:
                # Empty pandas dataframes behave weirdly here such that calling `m.df` raises:
                # ValueError: The truth value of a {type(self).__name__} is ambiguous. Use a.empty, a.bool(), a.item(),
                #   a.any() or a.all()
                # Since dataframes aren't registrable entities to begin with we swallow any errors they raise and
                # continue looping through m.
                _logging.warning("Caught ValueError {} while attempting to auto-assign name".format(err))
                pass

        _logging.error(f"Could not find LHS for {self} in {self._instantiated_in}")
        raise _system_exceptions.FlyteSystemException(f"Error looking for LHS in {self._instantiated_in}")


def isnested(func: Callable) -> bool:
    """
    Returns true if a function is local to another function and is not accessible through a module

    This would essentially be any function with a `.<local>.` (defined within a function) e.g.

    .. code:: python

        def foo():
            def foo_inner():
                pass
            pass

    In the above example `foo_inner` is the local function or a nested function.
    """
    return func.__code__.co_flags & inspect.CO_NESTED != 0


def istestfunction(func) -> bool:
    """
    Returns true if the function is defined in a test module. A test module has to have `test_` as the prefix.
    False in all other cases
    """
    mod = inspect.getmodule(func)
    if mod:
        mod_name = mod.__name__
        if "." in mod_name:
            mod_name = mod_name.split(".")[-1]
        return mod_name.startswith("test_")
    return False



"""
This file is for old style plugins - for new plugins that work with the Python native-typed Flytekit, please
refer to the plugin specific directory underneath the plugins folder at the top level of this repository.
"""
from flytekit.tools import lazy_loader as _lazy_loader

pyspark = _lazy_loader.lazy_load_module("pyspark")  # type: _lazy_loader._LazyLoadModule

k8s = _lazy_loader.lazy_load_module("k8s")  # type: _lazy_loader._LazyLoadModule
type(k8s).add_sub_module("io.api.core.v1.generated_pb2")
type(k8s).add_sub_module("io.apimachinery.pkg.api.resource.generated_pb2")

flyteidl = _lazy_loader.lazy_load_module("flyteidl")  # type: _lazy_loader._LazyLoadModule
type(flyteidl).add_sub_module("plugins.sidecar_pb2")

numpy = _lazy_loader.lazy_load_module("numpy")  # type: _lazy_loader._LazyLoadModule
pandas = _lazy_loader.lazy_load_module("pandas")  # type: _lazy_loader._LazyLoadModule

hmsclient = _lazy_loader.lazy_load_module("hmsclient")  # type: _lazy_loader._LazyLoadModule
type(hmsclient).add_sub_module("genthrift.hive_metastore.ttypes")

sagemaker_training = _lazy_loader.lazy_load_module("sagemaker_training")  # type: _lazy_loader._LazyLoadModule

papermill = _lazy_loader.lazy_load_module("papermill")  # type: _lazy_loader._LazyLoadModule

_lazy_loader.LazyLoadPlugin("spark", ["pyspark>=2.4.0,<3.0.0"], [pyspark])

_lazy_loader.LazyLoadPlugin("spark3", ["pyspark>=3.0.0"], [pyspark])

_lazy_loader.LazyLoadPlugin("sidecar", ["k8s-proto>=0.0.3,<1.0.0"], [k8s, flyteidl])

_lazy_loader.LazyLoadPlugin(
    "schema",
    ["numpy>=1.14.0,<2.0.0", "pandas>=0.22.0,<2.0.0", "pyarrow>=0.11.0,<1.0.0"],
    [numpy, pandas],
)

_lazy_loader.LazyLoadPlugin("hive_sensor", ["hmsclient>=0.0.1,<1.0.0"], [hmsclient])

_lazy_loader.LazyLoadPlugin("sagemaker", ["sagemaker-training>=3.6.2,<4.0.0"], [sagemaker_training])

_lazy_loader.LazyLoadPlugin("papermill", ["papermill>=2.0.0,<3.0.0"], [papermill])



from flytekit.common.exceptions import user as _user_exceptions
from flytekit.models.core import identifier as _core_identifier


class Identifier(_core_identifier.Identifier):

    _STRING_TO_TYPE_MAP = {
        "lp": _core_identifier.ResourceType.LAUNCH_PLAN,
        "wf": _core_identifier.ResourceType.WORKFLOW,
        "tsk": _core_identifier.ResourceType.TASK,
    }
    _TYPE_TO_STRING_MAP = {v: k for k, v in _STRING_TO_TYPE_MAP.items()}

    @classmethod
    def promote_from_model(cls, base_model: _core_identifier.Identifier) -> "Identifier":
        return cls(base_model.resource_type, base_model.project, base_model.domain, base_model.name, base_model.version)

    @classmethod
    def from_urn(cls, urn: str) -> "Identifier":
        """
        Parses a string urn in the correct format into an identifier
        """
        segments = urn.split(":")
        if len(segments) != 5:
            raise _user_exceptions.FlyteValueException(
                urn,
                "The provided string was not in a parseable format. The string for an identifier must be in the "
                "format entity_type:project:domain:name:version.",
            )

        resource_type, project, domain, name, version = segments

        if resource_type not in cls._STRING_TO_TYPE_MAP:
            raise _user_exceptions.FlyteValueException(
                resource_type,
                "The provided string could not be parsed. The first element of an identifier must be one of: "
                f"{list(cls._STRING_TO_TYPE_MAP.keys())}. ",
            )

        return cls(cls._STRING_TO_TYPE_MAP[resource_type], project, domain, name, version)

    def __str__(self):
        return (
            f"{type(self)._TYPE_TO_STRING_MAP.get(self.resource_type, '<unknown>')}:"
            f"{self.project}:"
            f"{self.domain}:"
            f"{self.name}:"
            f"{self.version}"
        )


class WorkflowExecutionIdentifier(_core_identifier.WorkflowExecutionIdentifier):
    @classmethod
    def promote_from_model(
        cls, base_model: _core_identifier.WorkflowExecutionIdentifier
    ) -> "WorkflowExecutionIdentifier":
        return cls(base_model.project, base_model.domain, base_model.name)

    @classmethod
    def from_urn(cls, string: str) -> "WorkflowExecutionIdentifier":
        """
        Parses a string in the correct format into an identifier
        """
        segments = string.split(":")
        if len(segments) != 4:
            raise _user_exceptions.FlyteValueException(
                string,
                "The provided string was not in a parseable format. The string for an identifier must be in the format"
                " ex:project:domain:name.",
            )

        resource_type, project, domain, name = segments

        if resource_type != "ex":
            raise _user_exceptions.FlyteValueException(
                resource_type,
                "The provided string could not be parsed. The first element of an execution identifier must be 'ex'.",
            )

        return cls(project, domain, name)

    def __str__(self):
        return f"ex:{self.project}:{self.domain}:{self.name}"


class TaskExecutionIdentifier(_core_identifier.TaskExecutionIdentifier):
    @classmethod
    def promote_from_model(cls, base_model: _core_identifier.TaskExecutionIdentifier) -> "TaskExecutionIdentifier":
        return cls(
            task_id=base_model.task_id,
            node_execution_id=base_model.node_execution_id,
            retry_attempt=base_model.retry_attempt,
        )

    @classmethod
    def from_urn(cls, string: str) -> "TaskExecutionIdentifier":
        """
        Parses a string in the correct format into an identifier
        """
        segments = string.split(":")
        if len(segments) != 10:
            raise _user_exceptions.FlyteValueException(
                string,
                "The provided string was not in a parseable format. The string for an identifier must be in the format"
                " te:exec_project:exec_domain:exec_name:node_id:task_project:task_domain:task_name:task_version:retry.",
            )

        resource_type, ep, ed, en, node_id, tp, td, tn, tv, retry = segments

        if resource_type != "te":
            raise _user_exceptions.FlyteValueException(
                resource_type,
                "The provided string could not be parsed. The first element of an execution identifier must be 'ex'.",
            )

        return cls(
            task_id=Identifier(_core_identifier.ResourceType.TASK, tp, td, tn, tv),
            node_execution_id=_core_identifier.NodeExecutionIdentifier(
                node_id=node_id,
                execution_id=_core_identifier.WorkflowExecutionIdentifier(ep, ed, en),
            ),
            retry_attempt=int(retry),
        )

    def __str__(self):
        return (
            "te:"
            f"{self.node_execution_id.execution_id.project}:"
            f"{self.node_execution_id.execution_id.domain}:"
            f"{self.node_execution_id.execution_id.name}:"
            f"{self.node_execution_id.node_id}:"
            f"{self.task_id.project}:"
            f"{self.task_id.domain}:"
            f"{self.task_id.name}:"
            f"{self.task_id.version}:"
            f"{self.retry_attempt}"
        )



import logging
import shlex as _schlex
import subprocess as _subprocess
import tempfile as _tempfile


def check_call(cmd_args, **kwargs):
    if not isinstance(cmd_args, list):
        cmd_args = _schlex.split(cmd_args)

    # Jupyter notebooks hijack I/O and thus we cannot dump directly to stdout.
    with _tempfile.TemporaryFile() as std_out:
        with _tempfile.TemporaryFile() as std_err:
            ret_code = _subprocess.Popen(cmd_args, stdout=std_out, stderr=std_err, **kwargs).wait()

            # Dump sub-process' std out into current std out
            std_out.seek(0)
            logging.info("Output of command '{}':\n{}\n".format(cmd_args, std_out.read()))

            if ret_code != 0:
                std_err.seek(0)
                err_str = std_err.read()
                logging.error("Error from command '{}':\n{}\n".format(cmd_args, err_str))

                raise Exception(
                    "Called process exited with error code: {}.  Stderr dump:\n\n{}".format(ret_code, err_str)
                )

    return 0



import six as _six
from flyteidl.core import compiler_pb2 as _compiler_pb2

from flytekit.models import common as _common
from flytekit.models.core import workflow as _core_workflow_models


class ConnectionSet(_common.FlyteIdlEntity):
    class IdList(_common.FlyteIdlEntity):
        def __init__(self, ids):
            """
            :param list[Text] ids:
            """
            self._ids = ids

        @property
        def ids(self):
            """
            :rtype: list[Text]
            """
            return self._ids

        def to_flyte_idl(self):
            """
            :rtype: flyteidl.core.compiler_pb2.ConnectionSet.IdList
            """
            return _compiler_pb2.ConnectionSet.IdList(ids=self.ids)

        @classmethod
        def from_flyte_idl(cls, p):
            """
            :param flyteidl.core.compiler_pb2.ConnectionSet.IdList p:
            :rtype: ConnectionSet.IdList
            """
            return cls(p.ids)

    def __init__(self, upstream, downstream):
        """
        :param dict[Text, ConnectionSet.IdList] upstream:
        :param dict[Text, ConnectionSet.IdList] downstream:
        """
        self._upstream = upstream
        self._downstream = downstream

    @property
    def upstream(self):
        """
        :rtype: dict[Text, ConnectionSet.IdList]
        """
        return self._upstream

    @property
    def downstream(self):
        """
        :rtype: dict[Text, ConnectionSet.IdList]
        """
        return self._downstream

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.compiler_pb2.ConnectionSet
        """
        return _compiler_pb2.ConnectionSet(
            upstream={k: v.to_flyte_idl() for k, v in _six.iteritems(self.upstream)},
            downstream={k: v.to_flyte_idl() for k, v in _six.iteritems(self.upstream)},
        )

    @classmethod
    def from_flyte_idl(cls, p):
        """
        :param flyteidl.core.compiler_pb2.ConnectionSet p:
        :rtype: ConnectionSet
        """
        return cls(
            upstream={k: ConnectionSet.IdList.from_flyte_idl(v) for k, v in _six.iteritems(p.upstream)},
            downstream={k: ConnectionSet.IdList.from_flyte_idl(v) for k, v in _six.iteritems(p.downstream)},
        )


class CompiledWorkflow(_common.FlyteIdlEntity):
    def __init__(self, template, connections):
        """
        :param flytekit.models.core.workflow.WorkflowTemplate template:
        :param ConnectionSet connections:
        """
        self._template = template
        self._connections = connections

    @property
    def template(self):
        """
        :rtype: flytekit.models.core.workflow.WorkflowTemplate
        """
        return self._template

    @property
    def connections(self):
        """
        :rtype: ConnectionSet
        """
        return self._connections

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.compiler_pb2.CompiledWorkflow
        """
        return _compiler_pb2.CompiledWorkflow(
            template=self.template.to_flyte_idl(),
            connections=self.connections.to_flyte_idl(),
        )

    @classmethod
    def from_flyte_idl(cls, p):
        """
        :param flyteidl.core.compiler_pb2.CompiledWorkflow p:
        :rtype: CompiledWorkflow
        """
        return cls(
            template=_core_workflow_models.WorkflowTemplate.from_flyte_idl(p.template),
            connections=ConnectionSet.from_flyte_idl(p.connections),
        )


# TODO: properly sort out the model code and remove one of these duplicate CompiledTasks
class CompiledTask(_common.FlyteIdlEntity):
    def __init__(self, template):
        """
        :param TODO template:
        """
        self._template = template

    @property
    def template(self):
        """
        :rtype: TODO
        """
        return self._template

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.compiler_pb2.CompiledTask
        """
        return _compiler_pb2.CompiledTask(template=self.template)  # TODO: .to_flyte_idl()

    @classmethod
    def from_flyte_idl(cls, p):
        """
        :param flyteidl.core.compiler_pb2.CompiledTask p:
        :rtype: CompiledTask
        """
        # TODO: Refactor task so we don't have cyclical import
        return cls(None)


class CompiledWorkflowClosure(_common.FlyteIdlEntity):
    def __init__(self, primary, sub_workflows, tasks):
        """
        :param CompiledWorkflow primary:
        :param list[CompiledWorkflow] sub_workflows:
        :param list[CompiledTask] tasks:
        """
        self._primary = primary
        self._sub_workflows = sub_workflows
        self._tasks = tasks

    @property
    def primary(self):
        """
        :rtype: CompiledWorkflow
        """
        return self._primary

    @property
    def sub_workflows(self):
        """
        :rtype: list[CompiledWorkflow]
        """
        return self._sub_workflows

    @property
    def tasks(self):
        """
        :rtype: list[CompiledTask]
        """
        return self._tasks

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.compiler_pb2.CompiledWorkflowClosure
        """
        return _compiler_pb2.CompiledWorkflowClosure(
            primary=self.primary.to_flyte_idl(),
            sub_workflows=[s.to_flyte_idl() for s in self.sub_workflows],
            tasks=[t.to_flyte_idl() for t in self.tasks],
        )

    @classmethod
    def from_flyte_idl(cls, p):
        """
        :param flyteidl.core.compiler_pb2.CompiledWorkflowClosure p:
        :rtype: CompiledWorkflowClosure
        """
        # This import is here to prevent a circular dependency issue.
        # TODO: properly sort out the model code and remove the duplicate CompiledTask
        from flytekit.models.task import CompiledTask as _CompiledTask

        return cls(
            primary=CompiledWorkflow.from_flyte_idl(p.primary),
            sub_workflows=[CompiledWorkflow.from_flyte_idl(s) for s in p.sub_workflows],
            tasks=[_CompiledTask.from_flyte_idl(t) for t in p.tasks],
        )



from flytekit.configuration import common as _config_common

DEFAULT_CPU_LIMIT = _config_common.FlyteStringConfigurationEntry("resources", "default_cpu_limit")
"""
If not specified explicitly when constructing a task, this limit will be applied as the default. Follows Kubernetes CPU
request/limit format.
"""

DEFAULT_CPU_REQUEST = _config_common.FlyteStringConfigurationEntry("resources", "default_cpu_request")
"""
If not specified explicitly when constructing a task, this request will be applied as the default. Follows Kubernetes
CPU request/limit format.
"""

DEFAULT_MEMORY_LIMIT = _config_common.FlyteStringConfigurationEntry("resources", "default_memory_limit")
"""
If not specified explicitly when constructing a task, this limit will be applied as the default. Follows Kubernetes
memory request/limit format.
"""

DEFAULT_MEMORY_REQUEST = _config_common.FlyteStringConfigurationEntry("resources", "default_memory_request")
"""
If not specified explicitly when constructing a task, this request will be applied as the default. Follows Kubernetes
memory request/limit format.
"""

DEFAULT_GPU_LIMIT = _config_common.FlyteStringConfigurationEntry("resources", "default_gpu_limit")
"""
If not specified explicitly when constructing a task, this limit will be applied as the default. Follows Kubernetes GPU
request/limit format.
"""

DEFAULT_GPU_REQUEST = _config_common.FlyteStringConfigurationEntry("resources", "default_gpu_request")
"""
If not specified explicitly when constructing a task, this request will be applied as the default. Follows Kubernetes
GPU request/limit format.
"""

DEFAULT_STORAGE_LIMIT = _config_common.FlyteStringConfigurationEntry("resources", "default_storage_limit")
"""
If not specified explicitly when constructing a task, this limit will be applied as the default. Follows Kubernetes
storage request/limit format.
"""

DEFAULT_STORAGE_REQUEST = _config_common.FlyteStringConfigurationEntry("resources", "default_storage_request")
"""
If not specified explicitly when constructing a task, this request will be applied as the default. Follows Kubernetes
storage request/limit format.
"""



from flyteidl.admin import workflow_pb2 as _admin_workflow

from flytekit.models import common as _common
from flytekit.models.core import compiler as _compiler_models
from flytekit.models.core import identifier as _identifier
from flytekit.models.core import workflow as _core_workflow


class WorkflowSpec(_common.FlyteIdlEntity):
    def __init__(self, template, sub_workflows):
        """
        This object fully encapsulates the specification of a workflow
        :param flytekit.models.core.workflow.WorkflowTemplate template:
        :param list[flytekit.models.core.workflow.WorkflowTemplate] sub_workflows:
        """
        self._template = template
        self._sub_workflows = sub_workflows

    @property
    def template(self):
        """
        :rtype: flytekit.models.core.workflow.WorkflowTemplate.WorkflowTemplate
        """
        return self._template

    @property
    def sub_workflows(self):
        """
        :rtype: list[flytekit.models.core.workflow.WorkflowTemplate.WorkflowTemplate]
        """
        return self._sub_workflows

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.workflow_pb2.WorkflowSpec
        """
        return _admin_workflow.WorkflowSpec(
            template=self._template.to_flyte_idl(),
            sub_workflows=[s.to_flyte_idl() for s in self._sub_workflows],
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param pb2_object: flyteidl.admin.workflow_pb2.WorkflowSpec
        :rtype: WorkflowSpec
        """
        return cls(
            _core_workflow.WorkflowTemplate.from_flyte_idl(pb2_object.template),
            [_core_workflow.WorkflowTemplate.from_flyte_idl(s) for s in pb2_object.sub_workflows],
        )


class Workflow(_common.FlyteIdlEntity):
    def __init__(self, id, closure):
        """
        :param flytekit.models.core.identifier.Identifier id:
        :param WorkflowClosure closure:
        """
        self._id = id
        self._closure = closure

    @property
    def id(self):
        """
        :rtype: flytekit.models.core.identifier.Identifier
        """
        return self._id

    @property
    def closure(self):
        """
        :rtype: WorkflowClosure
        """
        return self._closure

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.workflow_pb2.Workflow
        """
        return _admin_workflow.Workflow(id=self.id.to_flyte_idl(), closure=self.closure.to_flyte_idl())

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.admin.workflow_pb2.Workflow pb2_object:
        :return: Workflow
        """
        return cls(
            id=_identifier.Identifier.from_flyte_idl(pb2_object.id),
            closure=WorkflowClosure.from_flyte_idl(pb2_object.closure),
        )


class WorkflowClosure(_common.FlyteIdlEntity):
    def __init__(self, compiled_workflow):
        """
        :param flytekit.models.core.compiler.CompiledWorkflowClosure compiled_workflow:
        """
        self._compiled_workflow = compiled_workflow

    @property
    def compiled_workflow(self):
        """
        :rtype: flytekit.models.core.compiler.CompiledWorkflowClosure
        """
        return self._compiled_workflow

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.admin.workflow_pb2.WorkflowClosure
        """
        return _admin_workflow.WorkflowClosure(compiled_workflow=self.compiled_workflow.to_flyte_idl())

    @classmethod
    def from_flyte_idl(cls, p):
        """
        :param flyteidl.admin.workflow_pb2.WorkflowClosure p:
        :rtype: WorkflowClosure
        """
        return cls(compiled_workflow=_compiler_models.CompiledWorkflowClosure.from_flyte_idl(p.compiled_workflow))



import logging as _logging
import os as _os

import click
import six as _six

from flytekit.clis.helpers import construct_literal_map_from_parameter_map as _construct_literal_map_from_parameter_map
from flytekit.clis.sdk_in_container import constants as _constants
from flytekit.clis.sdk_in_container.constants import (
    CTX_DOMAIN,
    CTX_PROJECT,
    CTX_VERSION,
    domain_option,
    project_option,
    version_option,
)
from flytekit.common import utils as _utils
from flytekit.common.launch_plan import SdkLaunchPlan as _SdkLaunchPlan
from flytekit.configuration.internal import DOMAIN as _DOMAIN
from flytekit.configuration.internal import IMAGE as _IMAGE
from flytekit.configuration.internal import PROJECT as _PROJECT
from flytekit.configuration.internal import VERSION as _VERSION
from flytekit.configuration.internal import look_up_version_from_image_tag as _look_up_version_from_image_tag
from flytekit.models import launch_plan as _launch_plan_model
from flytekit.models.core import identifier as _identifier
from flytekit.tools.module_loader import iterate_registerable_entities_in_order


class LaunchPlanAbstractGroup(click.Group):
    """
    This class iterates over the workflow folders and loads all workflows that are implemented via the programming
    model.
    """

    def __init__(self, name, **attrs):
        super(LaunchPlanAbstractGroup, self).__init__(name, commands=None, **attrs)

    def list_commands(self, ctx):
        commands = []
        lps = {}
        pkgs = ctx.obj[_constants.CTX_PACKAGES]
        # Discover all launch plans by loading the modules
        for m, k, lp in iterate_registerable_entities_in_order(
            pkgs, include_entities={_SdkLaunchPlan}, detect_unreferenced_entities=False
        ):
            safe_name = _utils.fqdn(m.__name__, k, entity_type=lp.resource_type)
            commands.append(safe_name)
            lps[safe_name] = lp

        ctx.obj["lps"] = lps
        commands.sort()

        return commands

    def get_command(self, ctx, lp_argument):
        # Get the launch plan object in one of two ways. If get_command is being called by the list function
        # then it should have been cached in the context.
        # If we are actually running the command, then it won't have been cached and we'll have to load everything again
        launch_plan = None
        pkgs = ctx.obj[_constants.CTX_PACKAGES]

        if "lps" in ctx.obj:
            launch_plan = ctx.obj["lps"][lp_argument]
        else:
            for m, k, lp in iterate_registerable_entities_in_order(
                pkgs,
                include_entities={_SdkLaunchPlan},
                detect_unreferenced_entities=False,
            ):
                safe_name = _utils.fqdn(m.__name__, k, entity_type=lp.resource_type)
                if lp_argument == safe_name:
                    launch_plan = lp

        if launch_plan is None:
            raise Exception("Could not load launch plan {}".format(lp_argument))

        launch_plan._id = _identifier.Identifier(
            _identifier.ResourceType.LAUNCH_PLAN,
            ctx.obj[_constants.CTX_PROJECT],
            ctx.obj[_constants.CTX_DOMAIN],
            lp_argument,
            ctx.obj[_constants.CTX_VERSION],
        )
        return self._get_command(ctx, launch_plan, lp_argument)

    def _get_command(self, ctx, lp, cmd_name):
        """
        :param ctx:
        :param flytekit.common.launch_plan.SdkLaunchPlan lp:
        :rtype: click.Command
        """
        pass


class LaunchPlanExecuteGroup(LaunchPlanAbstractGroup):
    def _get_command(self, ctx, lp, cmd_name):
        """
        This function returns the function that click will actually use to execute a specific launch plan.  It also
        stores the launch plan python object and the command name in the closure.

        :param ctx:
        :param flytekit.common.launch_plan.SdkLaunchPlan lp:
        :param Text cmd_name: The name of the launch plan, as passed in from the abstract class
        """

        def _execute_lp(**kwargs):
            for input_name in _six.iterkeys(kwargs):
                if isinstance(kwargs[input_name], tuple):
                    kwargs[input_name] = list(kwargs[input_name])

            inputs = _construct_literal_map_from_parameter_map(lp.default_inputs, kwargs)
            execution = lp.execute_with_literals(
                ctx.obj[_constants.CTX_PROJECT],
                ctx.obj[_constants.CTX_DOMAIN],
                literal_inputs=inputs,
                notification_overrides=ctx.obj.get(_constants.CTX_NOTIFICATIONS, None),
            )
            click.echo(
                click.style(
                    "Workflow scheduled, execution_id={}".format(_six.text_type(execution.id)),
                    fg="blue",
                )
            )

        command = click.Command(name=cmd_name, callback=_execute_lp)

        # Iterate through the workflow's inputs
        for var_name in sorted(lp.default_inputs.parameters):
            param = lp.default_inputs.parameters[var_name]
            # TODO: Figure out how to better handle the fact that we want strings to parse,
            # but we probably shouldn't have click say that that's the type on the CLI.
            help_msg = "{} Type: {}".format(
                _six.text_type(param.var.description), _six.text_type(param.var.type)
            ).strip()

            if param.required:
                # If it's a required input, add the required flag
                wrapper = click.option(
                    "--{}".format(var_name),
                    required=True,
                    type=_six.text_type,
                    help=help_msg,
                )
            else:
                # If it's not a required input, it should have a default
                # Use to_python_std so that the text of the default ends up being parseable, if not, the click
                # arg would look something like 'Integer(10)'. If the user specified '11' on the cli, then
                # we'd get '11' and then we'd need annoying logic to differentiate between the default text
                # and user text.
                default = param.default.to_python_std()
                wrapper = click.option(
                    "--{}".format(var_name),
                    default="{}".format(_six.text_type(default)),
                    type=_six.text_type,
                    help="{}. Default: {}".format(help_msg, _six.text_type(default)),
                )

            command = wrapper(command)

        return command


@click.group("lp")
@project_option
@domain_option
@version_option
@click.pass_context
def launch_plans(ctx, project, domain, version):
    """
    Launch plan control group, including executions
    """

    version = version or _look_up_version_from_image_tag(_IMAGE.get())
    if not version:
        raise click.UsageError("Could not find image from config, please specify a value for ``--version``")

    ctx.obj[CTX_PROJECT] = project
    ctx.obj[CTX_DOMAIN] = domain
    ctx.obj[CTX_VERSION] = version
    _os.environ[_PROJECT.env_var] = project
    _os.environ[_DOMAIN.env_var] = domain
    _os.environ[_VERSION.env_var] = version


@click.group("execute", cls=LaunchPlanExecuteGroup)
@click.pass_context
def execute_launch_plan(ctx):
    """
    Execute launch plans found in this container
    """
    pass


def activate_all_impl(project, domain, version, pkgs, ignore_schedules=False):
    # TODO: This should be a transaction to ensure all or none are updated
    # TODO: We should optionally allow deactivation of missing launch plans

    # Discover all launch plans by loading the modules
    _logging.info(f"Setting this version's {version} launch plans active in {project} {domain}")
    for m, k, lp in iterate_registerable_entities_in_order(
        pkgs, include_entities={_SdkLaunchPlan}, detect_unreferenced_entities=False
    ):
        lp._id = _identifier.Identifier(
            _identifier.ResourceType.LAUNCH_PLAN,
            project,
            domain,
            _utils.fqdn(m.__name__, k, entity_type=lp.resource_type),
            version,
        )
        if not (lp.is_scheduled and ignore_schedules):
            _logging.info(f"Setting active {_utils.fqdn(m.__name__, k, entity_type=lp.resource_type)}")
            lp.update(_launch_plan_model.LaunchPlanState.ACTIVE)


@click.command("activate-all-schedules")
@click.option(
    "-v",
    "--version",
    type=str,
    help="Version to register tasks with. This is normally parsed from the" "image, but you can override here.",
)
@click.pass_context
def activate_all_schedules(ctx, version=None):
    """
    THIS COMMAND IS DEPRECATED. PLEASE USE activate-all

    The behavior of this command is identical to activate-all.
    """
    click.secho(
        "activate-all-schedules is deprecated, please use activate-all instead.",
        color="yellow",
    )
    project = ctx.obj[_constants.CTX_PROJECT]
    domain = ctx.obj[_constants.CTX_DOMAIN]
    pkgs = ctx.obj[_constants.CTX_PACKAGES]
    version = version or ctx.obj[_constants.CTX_VERSION] or _look_up_version_from_image_tag(_IMAGE.get())
    activate_all_impl(project, domain, version, pkgs)


@click.command("activate-all")
@click.option(
    "-v",
    "--version",
    type=str,
    help="Version to register tasks with. This is normally parsed from the" "image, but you can override here.",
)
@click.option(
    "--ignore-schedules",
    is_flag=True,
    help="Activate all except for launch plans with schedules.",
)
@click.pass_context
def activate_all(ctx, version=None, ignore_schedules=False):
    """
    This command will activate all found launch plans at the given version.  If there are existing
    active launch plans that collide on project, domain, and name, but differ on version, those will be
    deactivated in favor of the version specified in this command. If a launch plan is associated with a schedule,
    the schedule will also be deactivated or activated as appropriate.

    Note:
        1.  Currently, this is not a transaction.  Therefore, if the command fails, it is possible that some schedules
            have been updated.
        2.  If a launch plan is scheduled on an older version for a given project, domain, and name AND there is not a
            matching scheduled launch plan found when running this command, the existing schedule will remain active
            until it is manually disabled.
    """
    project = ctx.obj[_constants.CTX_PROJECT]
    domain = ctx.obj[_constants.CTX_DOMAIN]
    pkgs = ctx.obj[_constants.CTX_PACKAGES]
    version = version or ctx.obj[_constants.CTX_VERSION] or _look_up_version_from_image_tag(_IMAGE.get())
    activate_all_impl(project, domain, version, pkgs, ignore_schedules=ignore_schedules)


launch_plans.add_command(execute_launch_plan)
launch_plans.add_command(activate_all_schedules)
launch_plans.add_command(activate_all)



from __future__ import annotations

import os
import typing
from pathlib import Path

from flytekit.core.context_manager import FlyteContext
from flytekit.core.type_engine import TypeEngine, TypeTransformer
from flytekit.models import types as _type_models
from flytekit.models.core import types as _core_types
from flytekit.models.literals import Blob, BlobMetadata, Literal, Scalar
from flytekit.models.types import LiteralType

T = typing.TypeVar("T")


class FlyteDirectory(os.PathLike, typing.Generic[T]):
    """
    .. warning::

        This class should not be used on very large datasets, as merely listing the dataset will cause
        the entire dataset to be downloaded. Listing on S3 and other backend object stores is not consistent
        and we should not need data to be downloaded to list.

    Please first read through the comments on the :py:class:`flytekit.types.file.FlyteFile` class as the
    implementation here is similar.

    One thing to note is that the os.PathLike type that comes with Python was used as a stand-in for FlyteFile.
    That is, if a task returns an os.PathLike, Flyte takes that to mean FlyteFile. There is no easy way to
    distinguish an os.PathLike where the user means a File and where the user means a Directory. As such, if you
    want to use a directory, you must declare all types as FlyteDirectory. You'll still be able to return a string
    literal though instead of a full-fledged FlyteDirectory object assuming the str is a directory.

    Use cases as inputs ::

        def t1(in1: FlyteDirectory):
            ...

        def t1(in1: FlyteDirectory["svg"]):
            ...

    As outputs:

    The contents of this local directory will be uploaded to the Flyte store. ::

        return FlyteDirectory("/path/to/dir/")

        return FlyteDirectory["svg"]("/path/to/dir/", remote_path="s3://special/output/location")

    Similar to the FlyteFile example, if you give an already remote location, it will not be copied to Flyte's
    durable store, the uri will just be stored as is. ::

        return FlyteDirectory("s3://some/other/folder")

    Note if you write a path starting with http/s, if anything ever tries to read it (i.e. use the literal
    as an input, it'll fail because the http proxy doesn't know how to download whole directories.

    The format [] bit is still there because in Flyte, directories are stored as Blob Types also, just like files, and
    the Blob type has the format field. The difference in the type field is represented in the ``dimensionality``
    field in the ``BlobType``.
    """

    def __init__(self, path: str, downloader: typing.Callable = None, remote_directory=None):
        """
        :param path: The source path that users are expected to call open() on
        :param downloader: Optional function that can be passed that used to delay downloading of the actual fil
            until a user actually calls open().
        :param remote_directory: If the user wants to return something and also specify where it should be uploaded to.
        """

        def noop():
            ...

        self._path = path
        self._downloader = downloader or noop
        self._downloaded = False
        self._remote_directory = remote_directory
        self._remote_source = None

    def __fspath__(self):
        """
        This function should be called by os.listdir as well.
        """
        self._downloader()
        self._downloaded = True
        return self._path

    @classmethod
    def extension(cls) -> str:
        return ""

    def __class_getitem__(cls, item: typing.Type) -> typing.Type[FlyteDirectory]:
        if item is None:
            return cls
        item = str(item)
        item = item.strip().lstrip("~").lstrip(".")
        if item == "":
            return cls

        class _SpecificFormatDirectoryClass(FlyteDirectory):
            # Get the type engine to see this as kind of a generic
            __origin__ = FlyteDirectory

            @classmethod
            def extension(cls) -> str:
                return item

        return _SpecificFormatDirectoryClass

    @property
    def downloaded(self) -> bool:
        return self._downloaded

    @property
    def remote_directory(self) -> typing.Optional[str]:
        return self._remote_directory

    @property
    def path(self) -> str:
        return self._path

    @property
    def remote_source(self) -> str:
        """
        If this is an input to a task, and the original path is s3://something, flytekit will download the
        directory for the user. In case the user wants access to the original path, it will be here.
        """
        return self._remote_source

    def __repr__(self):
        return self._path

    def __str__(self):
        return self._path


class FlyteDirToMultipartBlobTransformer(TypeTransformer[FlyteDirectory]):
    """
    This transformer handles conversion between the Python native FlyteDirectory class defined above, and the Flyte
    IDL literal/type of Multipart Blob. Please see the FlyteDirectory comments for additional information.
    """

    def __init__(self):
        super().__init__(name="FlyteDirectory", t=FlyteDirectory)

    @staticmethod
    def get_format(t: typing.Type[FlyteDirectory]) -> str:
        return t.extension()

    @staticmethod
    def _blob_type(format: str) -> _core_types.BlobType:
        return _core_types.BlobType(format=format, dimensionality=_core_types.BlobType.BlobDimensionality.MULTIPART)

    def get_literal_type(self, t: typing.Type[FlyteDirectory]) -> LiteralType:
        return _type_models.LiteralType(blob=self._blob_type(format=FlyteDirToMultipartBlobTransformer.get_format(t)))

    def to_literal(
        self,
        ctx: FlyteContext,
        python_val: FlyteDirectory,
        python_type: typing.Type[FlyteDirectory],
        expected: LiteralType,
    ) -> Literal:

        remote_directory = None
        should_upload = True

        # There are two kinds of literals we handle, either an actual FlyteDirectory, or a string path to a directory.
        # Handle the FlyteDirectory case
        if isinstance(python_val, FlyteDirectory):
            # If the object has a remote source, then we just convert it back.
            if python_val._remote_source is not None:
                meta = BlobMetadata(type=self._blob_type(format=self.get_format(python_type)))
                return Literal(scalar=Scalar(blob=Blob(metadata=meta, uri=python_val._remote_source)))

            source_path = python_val.path
            if python_val.remote_directory is False:
                # If the user specified the remote_path to be False, that means no matter what, do not upload
                should_upload = False
            else:
                # Otherwise, if not an "" use the user-specified remote path instead of the random one
                remote_directory = python_val.remote_directory or None

        # Handle the string case
        else:
            if not (isinstance(python_val, os.PathLike) or isinstance(python_val, str)):
                raise AssertionError(f"Expected FlyteDirectory or os.PathLike object, received {type(python_val)}")

            source_path = python_val
            # Only do this check if it's a local directory.
            if not ctx.file_access.is_remote(source_path):
                p = Path(source_path)
                if not p.is_dir():
                    raise AssertionError(f"Expected a directory. {source_path} is not a directory")

        # For remote values, say s3://some/extant/dir/, we will not upload to Flyte's store (S3/GCS)
        # and just return a literal with a uri equal to the path given
        if ctx.file_access.is_remote(source_path) or not should_upload:
            meta = BlobMetadata(type=self._blob_type(format=self.get_format(python_type)))
            return Literal(scalar=Scalar(blob=Blob(metadata=meta, uri=source_path)))

        # For local paths, we will upload to the Flyte store (note that for local execution, the remote store is just
        # a subfolder), unless remote_path=False was given
        else:
            if remote_directory is None:
                remote_directory = ctx.file_access.get_random_remote_directory()
            ctx.file_access.put_data(source_path, remote_directory, is_multipart=True)
            meta = BlobMetadata(type=self._blob_type(format=self.get_format(python_type)))
            return Literal(scalar=Scalar(blob=Blob(metadata=meta, uri=remote_directory)))

    def to_python_value(
        self, ctx: FlyteContext, lv: Literal, expected_python_type: typing.Type[FlyteDirectory]
    ) -> FlyteDirectory:

        uri = lv.scalar.blob.uri

        # This is a local file path, like /usr/local/my_file, don't mess with it. Certainly, downloading it doesn't
        # make any sense.
        if not ctx.file_access.is_remote(uri):
            return expected_python_type(uri)

        # For the remote case, return an FlyteDirectory object that can download
        local_folder = ctx.file_access.get_random_local_directory()

        def _downloader():
            return ctx.file_access.get_data(uri, local_folder, is_multipart=True)

        expected_format = self.get_format(expected_python_type)

        fd = FlyteDirectory[expected_format](local_folder, _downloader)
        fd._remote_source = uri

        return fd


TypeEngine.register(FlyteDirToMultipartBlobTransformer())



from flytekit.configuration import common as _config_common

WORKFLOW_PACKAGES = _config_common.FlyteStringListConfigurationEntry("sdk", "workflow_packages", default=[])
"""
This is a comma-delimited list of packages that SDK tools will use to discover entities for the purpose of registration
and execution of entities.
"""

EXECUTION_ENGINE = _config_common.FlyteStringConfigurationEntry("sdk", "execution_engine", default="flyte")
"""
This is a comma-delimited list of package strings, in order, for resolving execution behavior.

TODO: Explain how this would be used to extend the SDK
"""

TYPE_ENGINES = _config_common.FlyteStringListConfigurationEntry("sdk", "type_engines", default=[])
"""
This is a comma-delimited list of package strings, in order, for resolving type behavior.

TODO: Explain how this would be used to extend the SDK
"""

LOCAL_SANDBOX = _config_common.FlyteStringConfigurationEntry("sdk", "local_sandbox", default="/tmp/flyte")
"""
This is the path where SDK will place files during local executions and testing.  The SDK will not automatically
clean up data in these directories.
"""

SDK_PYTHON_VENV = _config_common.FlyteStringListConfigurationEntry("sdk", "python_venv", default=[])
"""
This is a list of commands/args which will be prefixed to the entrypoint command by SDK.
"""

ROLE = _config_common.FlyteStringConfigurationEntry("sdk", "role")
"""
This is the role the SDK will use by default to execute workflows.  For example, in AWS this should be an IAM role
string.
"""

NAME_FORMAT = _config_common.FlyteStringConfigurationEntry("sdk", "name_format", default="{module}.{name}")
"""
This is a Python format string which the SDK will use to generate names for discovered entities.  The default is
'{module}.{name}' which will result in strings like 'package.module.name'.  Any template portion of the string can only
include 'module' or 'name'.  So '{name}' is valid, but '{key}' is not.
"""

TASK_NAME_FORMAT = _config_common.FlyteStringConfigurationEntry("sdk", "task_name_format", fallback=NAME_FORMAT)
"""
This is a Python format string which the SDK will use to generate names for tasks. Any template portion of the
string can only include 'module' or 'name'.  So '{name}' is valid, but '{key}' is not. If not specified,
we fall back to the configuration for :py:attr:`flytekit.configuration.sdk.NAME_FORMAT`
"""

WORKFLOW_NAME_FORMAT = _config_common.FlyteStringConfigurationEntry("sdk", "workflow_name_format", fallback=NAME_FORMAT)
"""
This is a Python format string which the SDK will use to generate names for workflows. Any template portion of the
string can only include 'module' or 'name'.  So '{name}' is valid, but '{key}' is not. If not specified,
we fall back to the configuration for :py:attr:`flytekit.configuration.sdk.NAME_FORMAT`
"""

LAUNCH_PLAN_NAME_FORMAT = _config_common.FlyteStringConfigurationEntry(
    "sdk", "launch_plan_name_format", fallback=NAME_FORMAT
)
"""
This is a Python format string which the SDK will use to generate names for launch plans. Any template portion of the
string can only include 'module' or 'name'.  So '{name}' is valid, but '{key}' is not. If not specified,
we fall back to the configuration for :py:attr:`flytekit.configuration.sdk.NAME_FORMAT`
"""

LOGGING_LEVEL = _config_common.FlyteIntegerConfigurationEntry("sdk", "logging_level", default=20)
"""
This is the default logging level for the Python logging library and will be set before user code runs.
Note that this configuration is special in that it is a runtime setting, not a compile time setting.  This is the only
runtime option in this file.
"""

PARQUET_ENGINE = _config_common.FlyteStringConfigurationEntry("sdk", "parquet_engine", default="pyarrow")
"""
This is the parquet engine to use when reading data from parquet files.
"""

FAST_REGISTRATION_DIR = _config_common.FlyteStringConfigurationEntry("sdk", "fast_registration_dir")
"""
This is the remote directory where fast-registered code will be uploaded to.
Users calling fast-execute need write permission to this directory.
Furthermore, it is important that whichever role executes your workflow has read access to this directory.
"""



from collections import OrderedDict

from flytekit.common.translator import get_serializable
from flytekit.core import context_manager
from flytekit.core.context_manager import Image, ImageConfig
from plugins.tests.sqlalchemy.test_task import tk as not_tk


def test_sql_lhs():
    assert not_tk.lhs == "tk"


def test_sql_command():
    default_img = Image(name="default", fqn="test", tag="tag")
    serialization_settings = context_manager.SerializationSettings(
        project="project",
        domain="domain",
        version="version",
        env=None,
        image_config=ImageConfig(default_image=default_img, images=[default_img]),
    )
    srz_t = get_serializable(OrderedDict(), serialization_settings, not_tk)
    assert srz_t.container.args[-7:] == [
        "--resolver",
        "flytekit.core.python_auto_container.default_task_resolver",
        "--",
        "task-module",
        "plugins.tests.sqlalchemy.test_task",
        "task-name",
        "tk",
    ]



from datetime import datetime as _datetime

import pytz as _pytz
import six as _six
from flyteidl.core import literals_pb2 as _literals_pb2
from google.protobuf.struct_pb2 import Struct

from flytekit.common.exceptions import user as _user_exceptions
from flytekit.models import common as _common
from flytekit.models.core import types as _core_types
from flytekit.models.types import OutputReference as _OutputReference
from flytekit.models.types import SchemaType as _SchemaType


class RetryStrategy(_common.FlyteIdlEntity):
    def __init__(self, retries):
        """
        :param int retries: Number of retries to attempt on recoverable failures.  If retries is 0, then
            only one attempt will be made.
        """
        self._retries = retries

    @property
    def retries(self):
        """
        Number of retries to attempt on recoverable failures.  If retries is 0, then only one attempt will be made.
        :rtype: int
        """
        return self._retries

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.RetryStrategy
        """
        return _literals_pb2.RetryStrategy(retries=self.retries)

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.literals_pb2.RetryStrategy pb2_object:
        :rtype: RetryStrategy
        """
        return cls(retries=pb2_object.retries)


class Primitive(_common.FlyteIdlEntity):
    def __init__(
        self,
        integer=None,
        float_value=None,
        string_value=None,
        boolean=None,
        datetime=None,
        duration=None,
    ):
        """
        This object proxies the primitives supported by the Flyte IDL system.  Only one value can be set.
        :param int integer: [Optional]
        :param float float_value: [Optional]
        :param Text string_value: [Optional]
        :param bool boolean: [Optional]
        :param datetime.timestamp datetime: [Optional]
        :param datetime.timedelta duration: [Optional]
        """
        self._integer = integer
        self._float_value = float_value
        self._string_value = string_value
        self._boolean = boolean
        if datetime is None:
            self._datetime = None
        elif isinstance(datetime, _datetime):
            self._datetime = datetime
        else:  # TODO Check for timestamp type?
            self._datetime = _datetime.utcfromtimestamp(datetime.seconds)
        self._duration = duration

    @property
    def integer(self):
        """
        :rtype: int
        """
        return self._integer

    @property
    def float_value(self):
        """
        :rtype: float
        """
        return self._float_value

    @property
    def string_value(self):
        """
        :rtype: Text
        """
        return self._string_value

    @property
    def boolean(self):
        """
        :rtype: bool
        """
        return self._boolean

    @property
    def datetime(self):
        """
        :rtype: datetime.datetime
        """
        if self._datetime is None or self._datetime.tzinfo is not None:
            return self._datetime
        return self._datetime.replace(tzinfo=_pytz.UTC)

    @property
    def duration(self):
        """
        :rtype: datetime.timedelta
        """
        return self._duration

    @property
    def value(self):
        """
        This returns whichever field is set.
        :rtype: T
        """
        for value in [
            self.integer,
            self.float_value,
            self.string_value,
            self.boolean,
            self.datetime,
            self.duration,
        ]:
            if value is not None:
                return value

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.Primitive
        """
        primitive = _literals_pb2.Primitive(
            integer=self.integer,
            float_value=self.float_value,
            string_value=self.string_value,
            boolean=self.boolean,
        )
        if self.datetime is not None:
            # Convert to UTC and remove timezone so protobuf behaves.
            primitive.datetime.FromDatetime(self.datetime.astimezone(_pytz.UTC).replace(tzinfo=None))
        if self.duration is not None:
            primitive.duration.FromTimedelta(self.duration)
        return primitive

    @classmethod
    def from_flyte_idl(cls, proto):
        """
        :param flyteidl.core.literals_pb2.Primitive proto:
        :rtype: Primitive
        """
        return cls(
            integer=proto.integer if proto.HasField("integer") else None,
            float_value=proto.float_value if proto.HasField("float_value") else None,
            string_value=proto.string_value if proto.HasField("string_value") else None,
            boolean=proto.boolean if proto.HasField("boolean") else None,
            datetime=proto.datetime.ToDatetime().replace(tzinfo=_pytz.UTC) if proto.HasField("datetime") else None,
            duration=proto.duration.ToTimedelta() if proto.HasField("duration") else None,
        )


class Binary(_common.FlyteIdlEntity):
    def __init__(self, value, tag):
        """
        :param bytes value:
        :param Text tag:
        """
        self._value = value
        self._tag = tag

    @property
    def value(self):
        """
        :rtype: bytes
        """
        return self._value

    @property
    def tag(self):
        """
        :rtype: Text
        """
        return self._tag

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.Binary
        """
        return _literals_pb2.Binary(value=self.value, tag=self.tag)

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.literals_pb2.Binary pb2_object:
        :rtype: Binary
        """
        return cls(value=pb2_object.value, tag=pb2_object.tag)


class BlobMetadata(_common.FlyteIdlEntity):
    def __init__(self, type):
        """
        :param flytekit.models.core.types.BlobType type: The type of the underlying blob
        """
        self._type = type

    @property
    def type(self):
        """
        :rtype: flytekit.models.core.types.BlobType
        """
        return self._type

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.BlobMetadata
        """
        return _literals_pb2.BlobMetadata(type=self.type.to_flyte_idl())

    @classmethod
    def from_flyte_idl(cls, proto):
        """
        :param flyteidl.core.literals_pb2.BlobMetadata proto:
        :rtype: BlobMetadata
        """
        return cls(type=_core_types.BlobType.from_flyte_idl(proto.type))


class Blob(_common.FlyteIdlEntity):
    def __init__(self, metadata, uri):
        """
        :param BlobMetadata metadata:
        :param Text uri: The location of this blob
        """
        self._metadata = metadata
        self._uri = uri

    @property
    def uri(self):
        """
        :rtype: Text
        """
        return self._uri

    @property
    def metadata(self):
        """
        :rtype: BlobMetadata
        """
        return self._metadata

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.Blob
        """
        return _literals_pb2.Blob(metadata=self.metadata.to_flyte_idl(), uri=self.uri)

    @classmethod
    def from_flyte_idl(cls, proto):
        """
        :param flyteidl.core.literals_pb2.Blob proto:
        :rtype: Blob
        """
        return cls(metadata=BlobMetadata.from_flyte_idl(proto.metadata), uri=proto.uri)


class Void(_common.FlyteIdlEntity):
    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.Void
        """
        return _literals_pb2.Void()

    @classmethod
    def from_flyte_idl(cls, proto):
        """
        :param flyteidl.core.literals_pb2.Void proto:
        :rtype: Void
        """
        return cls()


class BindingDataMap(_common.FlyteIdlEntity):
    def __init__(self, bindings):
        """
        A map of BindingData items.  Can be a recursive structure

        :param dict[string, BindingData] bindings: Map of strings to Bindings
        """
        self._bindings = bindings

    @property
    def bindings(self):
        """
        Map of strings to Bindings
        :rtype: dict[string, BindingData]
        """
        return self._bindings

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.BindingDataMap
        """
        return _literals_pb2.BindingDataMap(bindings={k: v.to_flyte_idl() for (k, v) in _six.iteritems(self.bindings)})

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.literals_pb2.BindingDataMap pb2_object:
        :rtype: flytekit.models.literals.BindingDataMap
        """

        return cls({k: BindingData.from_flyte_idl(v) for (k, v) in _six.iteritems(pb2_object.bindings)})


class BindingDataCollection(_common.FlyteIdlEntity):
    def __init__(self, bindings):
        """
        A list of BindingData items.

        :param list[BindingData] bindings:
        """
        self._bindings = bindings

    @property
    def bindings(self):
        """
        :rtype: list[BindingData]
        """
        return self._bindings

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.BindingDataCollection
        """
        return _literals_pb2.BindingDataCollection(bindings=[b.to_flyte_idl() for b in self.bindings])

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.literals_pb2.BindingDataCollection pb2_object:
        :rtype: flytekit.models.literals.BindingDataCollection
        """
        return cls([BindingData.from_flyte_idl(b) for b in pb2_object.bindings])


class BindingData(_common.FlyteIdlEntity):
    def __init__(self, scalar=None, collection=None, promise=None, map=None):
        """
        Specifies either a simple value or a reference to another output. Only one of the input arguments may be
        specified.

        :param Scalar scalar: [Optional] A simple scalar value.
        :param BindingDataCollection collection: [Optional] A collection of binding data. This allows nesting of
            binding data to any number of levels.
        :param flytekit.models.types.OutputReference promise: [Optional] References an output promised by another node.
        :param BindingDataMap map: [Optional] A map of bindings. The key is always a string.
        """
        self._scalar = scalar
        self._collection = collection
        self._promise = promise
        self._map = map

    @property
    def scalar(self):
        """
        A simple scalar value.
        :rtype: Scalar
        """
        return self._scalar

    @property
    def collection(self):
        """
        [Optional] A collection of binding data. This allows nesting of binding data to any number of levels.
        :rtype: BindingDataCollection
        """
        return self._collection

    @property
    def promise(self):
        """
        [Optional] References an output promised by another node.
        :rtype: flytekit.models.types.OutputReference
        """
        return self._promise

    @property
    def map(self):
        """
        [Optional] A map of bindings. The key is always a string.
        :rtype: BindingDataMap
        """
        return self._map

    @property
    def value(self):
        """
        Returns whichever value is set
        :rtype: T
        """
        return self.scalar or self.collection or self.promise or self.map

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.BindingData
        """
        return _literals_pb2.BindingData(
            scalar=self.scalar.to_flyte_idl() if self.scalar is not None else None,
            collection=self.collection.to_flyte_idl() if self.collection is not None else None,
            promise=self.promise.to_flyte_idl() if self.promise is not None else None,
            map=self.map.to_flyte_idl() if self.map is not None else None,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.literals_pb2.BindingData pb2_object:
        :return: BindingData
        """
        return cls(
            scalar=Scalar.from_flyte_idl(pb2_object.scalar) if pb2_object.HasField("scalar") else None,
            collection=BindingDataCollection.from_flyte_idl(pb2_object.collection)
            if pb2_object.HasField("collection")
            else None,
            promise=_OutputReference.from_flyte_idl(pb2_object.promise) if pb2_object.HasField("promise") else None,
            map=BindingDataMap.from_flyte_idl(pb2_object.map) if pb2_object.HasField("map") else None,
        )

    def to_literal_model(self):
        """
        Converts current binding data into a Literal asserting that there are no promises in the bindings.
        :rtype: Literal
        """
        if self.promise:
            raise _user_exceptions.FlyteValueException(
                self.promise,
                "Cannot convert BindingData to a Literal because " "it has a promise.",
            )
        elif self.scalar:
            return Literal(scalar=self.scalar)
        elif self.collection:
            return Literal(
                collection=LiteralCollection(
                    literals=[binding.to_literal_model() for binding in self.collection.bindings]
                )
            )
        elif self.map:
            return Literal(map=LiteralMap(literals={k: binding.to_literal_model() for k, binding in self.map.bindings}))


class Binding(_common.FlyteIdlEntity):
    def __init__(self, var, binding):
        """
        An input/output binding of a variable to either static value or a node output.

        :param Text var: A variable name, must match an input or output variable of the node.
        :param BindingData binding: Data to use to bind this variable.
        """
        self._var = var
        self._binding = binding

    @property
    def var(self):
        """
        A variable name, must match an input or output variable of the node.
        :rtype: Text
        """
        return self._var

    @property
    def binding(self):
        """
        Data to use to bind this variable.
        :rtype: BindingData
        """
        return self._binding

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.Binding
        """
        return _literals_pb2.Binding(var=self.var, binding=self.binding.to_flyte_idl())

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.literals_pb2.Binding pb2_object:
        :rtype: flytekit.core.models.literals.Binding
        """
        return cls(pb2_object.var, BindingData.from_flyte_idl(pb2_object.binding))


class Schema(_common.FlyteIdlEntity):
    def __init__(self, uri, type):
        """
        A strongly typed schema that defines the interface of data retrieved from the underlying storage medium.

        :param Text uri:
        :param flytekit.models.types.SchemaType type:
        """
        self._uri = uri
        self._type = type

    @property
    def uri(self):
        """
        :rtype: Text
        """
        return self._uri

    @property
    def type(self):
        """
        :rtype: flytekit.models.types.SchemaType
        """
        return self._type

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.Schema
        """
        return _literals_pb2.Schema(uri=self.uri, type=self.type.to_flyte_idl())

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.literals_pb2.Schema pb2_object:
        :rtype: Schema
        """
        return cls(uri=pb2_object.uri, type=_SchemaType.from_flyte_idl(pb2_object.type))


class LiteralCollection(_common.FlyteIdlEntity):
    def __init__(self, literals):
        """
        :param list[Literal] literals: underlying list of literals in this collection.
        """
        self._literals = literals

    @property
    def literals(self):
        """
        :rtype: list[Literal]
        """
        return self._literals

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.LiteralCollection
        """
        return _literals_pb2.LiteralCollection(literals=[l.to_flyte_idl() for l in self.literals])

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.literals_pb2.LiteralCollection pb2_object:
        :rtype: LiteralCollection
        """
        return cls([Literal.from_flyte_idl(l) for l in pb2_object.literals])


class LiteralMap(_common.FlyteIdlEntity):
    def __init__(self, literals):
        """
        :param dict[Text, Literal] literals: A dictionary mapping Text key names to Literal objects.
        """
        self._literals = literals

    @property
    def literals(self):
        """
        A dictionary mapping Text key names to Literal objects.
        :rtype: dict[Text, Literal]
        """
        return self._literals

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.LiteralMap
        """
        return _literals_pb2.LiteralMap(literals={k: v.to_flyte_idl() for k, v in _six.iteritems(self.literals)})

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.literals_pb2.LiteralMap pb2_object:
        :rtype: LiteralMap
        """
        return cls({k: Literal.from_flyte_idl(v) for k, v in _six.iteritems(pb2_object.literals)})


class Scalar(_common.FlyteIdlEntity):
    def __init__(
        self,
        primitive: Primitive = None,
        blob: Blob = None,
        binary: Binary = None,
        schema: Schema = None,
        none_type: Void = None,
        error=None,
        generic: Struct = None,
    ):
        """
        Scalar wrapper around Flyte types.  Only one can be specified.

        :param Primitive primitive:
        :param Blob blob:
        :param Binary binary:
        :param Schema schema:
        :param Void none_type:
        :param error:
        :param google.protobuf.struct_pb2.Struct generic:
        """

        self._primitive = primitive
        self._blob = blob
        self._binary = binary
        self._schema = schema
        self._none_type = none_type
        self._error = error
        self._generic = generic

    @property
    def primitive(self):
        """
        :rtype: Primitive
        """
        return self._primitive

    @property
    def blob(self):
        """
        :rtype: Blob
        """
        return self._blob

    @property
    def binary(self):
        """
        :rtype: Binary
        """
        return self._binary

    @property
    def schema(self):
        """
        :rtype: Schema
        """
        return self._schema

    @property
    def none_type(self):
        """
        :rtype: Void
        """
        return self._none_type

    @property
    def error(self):
        """
        :rtype: TODO
        """
        return self._error

    @property
    def generic(self):
        """
        :rtype: google.protobuf.struct_pb2.Struct
        """
        return self._generic

    @property
    def value(self):
        """
        Returns whichever value is set
        :rtype: T
        """
        return self.primitive or self.blob or self.binary or self.schema or self.none_type or self.error

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.Scalar
        """
        return _literals_pb2.Scalar(
            primitive=self.primitive.to_flyte_idl() if self.primitive is not None else None,
            blob=self.blob.to_flyte_idl() if self.blob is not None else None,
            binary=self.binary.to_flyte_idl() if self.binary is not None else None,
            schema=self.schema.to_flyte_idl() if self.schema is not None else None,
            none_type=self.none_type.to_flyte_idl() if self.none_type is not None else None,
            error=self.error if self.error is not None else None,
            generic=self.generic,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.literals_pb2.Scalar pb2_object:
        :rtype: flytekit.models.literals.Scalar
        """
        # todo finish
        return cls(
            primitive=Primitive.from_flyte_idl(pb2_object.primitive) if pb2_object.HasField("primitive") else None,
            blob=Blob.from_flyte_idl(pb2_object.blob) if pb2_object.HasField("blob") else None,
            binary=Binary.from_flyte_idl(pb2_object.binary) if pb2_object.HasField("binary") else None,
            schema=Schema.from_flyte_idl(pb2_object.schema) if pb2_object.HasField("schema") else None,
            none_type=Void.from_flyte_idl(pb2_object.none_type) if pb2_object.HasField("none_type") else None,
            error=pb2_object.error if pb2_object.HasField("error") else None,
            generic=pb2_object.generic if pb2_object.HasField("generic") else None,
        )


class Literal(_common.FlyteIdlEntity):
    def __init__(self, scalar: Scalar = None, collection: LiteralCollection = None, map: LiteralMap = None):
        """
        :param Scalar scalar:
        :param LiteralCollection collection:
        :param LiteralMap map:
        """
        self._scalar = scalar
        self._collection = collection
        self._map = map

    @property
    def scalar(self):
        """
        If not None, this value holds a scalar value which can be further unpacked.
        :rtype: Scalar
        """
        return self._scalar

    @property
    def collection(self):
        """
        If not None, this value holds a collection of Literal values which can be further unpacked.
        :rtype: LiteralCollection
        """
        return self._collection

    @property
    def map(self):
        """
        If not None, this value holds a map of Literal values which can be further unpacked.
        :rtype: LiteralMap
        """
        return self._map

    @property
    def value(self):
        """
        Returns one of the scalar, collection, or map properties based on which one is set.
        :rtype: T
        """
        return self.scalar or self.collection or self.map

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.literals_pb2.Literal
        """
        return _literals_pb2.Literal(
            scalar=self.scalar.to_flyte_idl() if self.scalar is not None else None,
            collection=self.collection.to_flyte_idl() if self.collection is not None else None,
            map=self.map.to_flyte_idl() if self.map is not None else None,
        )

    @classmethod
    def from_flyte_idl(cls, pb2_object):
        """
        :param flyteidl.core.literals_pb2.Literal pb2_object:
        :rtype: Literal
        """
        collection = None
        if pb2_object.HasField("collection"):
            collection = LiteralCollection.from_flyte_idl(pb2_object.collection)

        return cls(
            scalar=Scalar.from_flyte_idl(pb2_object.scalar) if pb2_object.HasField("scalar") else None,
            collection=collection,
            map=LiteralMap.from_flyte_idl(pb2_object.map) if pb2_object.HasField("map") else None,
        )



from flytekit.common import constants as _constants
from flytekit.configuration import common as _config_common

URL = _config_common.FlyteStringConfigurationEntry("platform", "url")

HTTP_URL = _config_common.FlyteStringConfigurationEntry("platform", "http_url", default=None)
"""
If not starting with either http or https, this setting should begin with // as per the urlparse library and
https://tools.ietf.org/html/rfc1808.html, otherwise the netloc will not be properly parsed.

Currently the only use-case for this configuration setting is for Auth discovery. This setting supports the case where
Flyte Admin's gRPC and HTTP points are deployed on different ports.
"""

INSECURE = _config_common.FlyteBoolConfigurationEntry("platform", "insecure", default=False)

CLOUD_PROVIDER = _config_common.FlyteStringConfigurationEntry(
    "platform", "cloud_provider", default=_constants.CloudProvider.AWS
)

AUTH = _config_common.FlyteBoolConfigurationEntry("platform", "auth", default=False)
"""
This config setting should not normally be filled in. Whether or not an admin server requires authentication should be
something published by the admin server itself (typically by returning a 401). However, to help with migration, this
config object is here to force the SDK to attempt the auth flow even without prompting by Admin.
"""



import os
import shutil
import tempfile

import doltcli as dolt
import pandas
import pytest
from flytekitplugins.dolt.schema import DoltConfig, DoltTable

from flytekit import task, workflow


@pytest.fixture(scope="function")
def doltdb_path():
    d = tempfile.TemporaryDirectory()
    try:
        db_path = os.path.join(d.name, "foo")
        yield db_path
    finally:
        shutil.rmtree(d.name)


@pytest.fixture(scope="function")
def dolt_config(doltdb_path):
    yield DoltConfig(
        db_path=doltdb_path,
        tablename="foo",
    )


@pytest.fixture(scope="function")
def db(doltdb_path):
    try:
        db = dolt.Dolt.init(doltdb_path)
        db.sql("create table bar (name text primary key, count bigint)")
        db.sql("insert into bar values ('Dilly', 3)")
        db.sql("select dolt_commit('-am', 'Initialize bar table')")
        yield db
    finally:
        pass


def test_dolt_table_write(db, dolt_config):
    @task
    def my_dolt(a: int) -> DoltTable:
        df = pandas.DataFrame([("Alice", a)], columns=["name", "count"])
        return DoltTable(data=df, config=dolt_config)

    @workflow
    def my_wf(a: int) -> DoltTable:
        return my_dolt(a=a)

    x = my_wf(a=5)
    assert x
    assert (x.data == pandas.DataFrame([("Alice", 5)], columns=["name", "count"])).all().all()


def test_dolt_table_read(db, dolt_config):
    @task
    def my_dolt(t: DoltTable) -> str:
        df = t.data
        return df.name.values[0]

    @workflow
    def my_wf(t: DoltTable) -> str:
        return my_dolt(t=t)

    dolt_config.tablename = "bar"
    x = my_wf(t=DoltTable(config=dolt_config))
    assert x == "Dilly"


def test_dolt_table_read_task_config(db, dolt_config):
    @task
    def my_dolt(t: DoltTable) -> str:
        df = t.data
        return df.name.values[0]

    @task
    def my_table() -> DoltTable:
        dolt_config.tablename = "bar"
        t = DoltTable(config=dolt_config)
        return t

    @workflow
    def my_wf() -> str:
        t = my_table()
        return my_dolt(t=t)

    x = my_wf()
    assert x == "Dilly"


def test_dolt_table_read_mixed_config(db, dolt_config):
    @task
    def my_dolt(t: DoltTable) -> str:
        df = t.data
        return df.name.values[0]

    @task
    def my_table(conf: DoltConfig) -> DoltTable:
        return DoltTable(config=conf)

    @workflow
    def my_wf(conf: DoltConfig) -> str:
        t = my_table(conf=conf)
        return my_dolt(t=t)

    dolt_config.tablename = "bar"
    x = my_wf(conf=dolt_config)

    assert x == "Dilly"


def test_dolt_sql_read(db, dolt_config):
    @task
    def my_dolt(t: DoltTable) -> str:
        df = t.data
        return df.name.values[0]

    @workflow
    def my_wf(t: DoltTable) -> str:
        return my_dolt(t=t)

    dolt_config.tablename = None
    dolt_config.sql = "select * from bar"
    x = my_wf(t=DoltTable(config=dolt_config))
    assert x == "Dilly"



from sys import exc_info as _exc_info
from traceback import format_tb as _format_tb

from six import reraise as _reraise
from wrapt import decorator as _decorator

from flytekit.common.exceptions import base as _base_exceptions
from flytekit.common.exceptions import system as _system_exceptions
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.models.core import errors as _error_model


class FlyteScopedException(Exception):
    def __init__(self, context, exc_type, exc_value, exc_tb, top_trim=0, bottom_trim=0, kind=None):
        self._exc_type = exc_type
        self._exc_value = exc_value
        self._exc_tb = exc_tb
        self._top_trim = top_trim
        self._bottom_trim = bottom_trim
        self._context = context
        self._kind = kind
        super(FlyteScopedException, self).__init__(str(self.value))

    @property
    def verbose_message(self):
        tb = self.traceback
        to_trim = self._top_trim
        while to_trim > 0 and tb.tb_next is not None:
            tb = tb.tb_next

        top_tb = tb
        limit = 0
        while tb is not None:
            limit += 1
            tb = tb.tb_next
        limit = max(0, limit - self._bottom_trim)

        lines = _format_tb(top_tb, limit=limit)
        lines = [line.rstrip() for line in lines]
        lines = "\n".join(lines).split("\n")
        traceback_str = "\n    ".join([""] + lines)

        format_str = "Traceback (most recent call last):\n" "{traceback}\n" "\n" "Message:\n" "\n" "    {message}"
        return format_str.format(traceback=traceback_str, message=str(self.value))

    def __str__(self):
        return str(self.value)

    @property
    def value(self):
        if isinstance(self._exc_value, FlyteScopedException):
            return self._exc_value.value
        return self._exc_value

    @property
    def traceback(self):
        if isinstance(self._exc_value, FlyteScopedException):
            return self._exc_value.traceback
        return self._exc_tb

    @property
    def type(self):
        if isinstance(self._exc_value, FlyteScopedException):
            return self._exc_value.type
        return self._exc_type

    @property
    def error_code(self):
        """
        :rtype: Text
        """
        if isinstance(self._exc_value, FlyteScopedException):
            return self._exc_value.error_code

        if hasattr(type(self._exc_value), "error_code"):
            return type(self._exc_value).error_code
        return "{}:Unknown".format(self._context)

    @property
    def kind(self):
        """
        :rtype: int
        """
        if self._kind is not None:
            # If kind is overriden, return it.
            return self._kind
        elif isinstance(self._exc_value, FlyteScopedException):
            # Otherwise, go lower in the scope to find the kind of exception.
            return self._exc_value.kind
        elif isinstance(self._exc_value, _base_exceptions.FlyteRecoverableException):
            # If it is an exception that is recoverable, we return it as such.
            return _error_model.ContainerError.Kind.RECOVERABLE
        else:
            # The remaining exceptions are considered unrecoverable.
            return _error_model.ContainerError.Kind.NON_RECOVERABLE


class FlyteScopedSystemException(FlyteScopedException):
    def __init__(self, exc_type, exc_value, exc_tb, **kwargs):
        super(FlyteScopedSystemException, self).__init__("SYSTEM", exc_type, exc_value, exc_tb, **kwargs)

    @property
    def verbose_message(self):
        """
        :rtype: Text
        """
        base_msg = super(FlyteScopedSystemException, self).verbose_message
        base_msg += "\n\nSYSTEM ERROR! Contact platform administrators."
        return base_msg


class FlyteScopedUserException(FlyteScopedException):
    def __init__(self, exc_type, exc_value, exc_tb, **kwargs):
        super(FlyteScopedUserException, self).__init__("USER", exc_type, exc_value, exc_tb, **kwargs)

    @property
    def verbose_message(self):
        """
        :rtype: Text
        """
        base_msg = super(FlyteScopedUserException, self).verbose_message
        base_msg += "\n\nUser error."
        return base_msg


_NULL_CONTEXT = 0
_USER_CONTEXT = 1
_SYSTEM_CONTEXT = 2

# Keep the stack with a null-context so we never have to range check when peeking back.
_CONTEXT_STACK = [_NULL_CONTEXT]


def _is_base_context():
    return _CONTEXT_STACK[-2] == _NULL_CONTEXT


@_decorator
def system_entry_point(wrapped, instance, args, kwargs):
    """
    Decorator for wrapping functions that enter a system context.  This should decorate every method a user might
    call.  This will allow us to add differentiation between what is a user error and what is a system failure.
    Furthermore, we will clean the exception trace so as to make more sense to the user--allowing them to know if they
    should take action themselves or pass on to the platform owners.  We will dispatch metrics and such appropriately.
    """
    try:
        _CONTEXT_STACK.append(_SYSTEM_CONTEXT)
        if _is_base_context():
            try:
                return wrapped(*args, **kwargs)
            except FlyteScopedException as ex:
                _reraise(ex.type, ex.value, ex.traceback)
        else:
            try:
                return wrapped(*args, **kwargs)
            except FlyteScopedException:
                # Just pass-on the exception that is already wrapped and scoped
                _reraise(*_exc_info())
            except _user_exceptions.FlyteUserException:
                # Re-raise from here.
                _reraise(
                    FlyteScopedUserException,
                    FlyteScopedUserException(*_exc_info()),
                    _exc_info()[2],
                )
            except Exception:
                # System error, raise full stack-trace all the way up the chain.
                _reraise(
                    FlyteScopedSystemException,
                    FlyteScopedSystemException(*_exc_info(), kind=_error_model.ContainerError.Kind.RECOVERABLE),
                    _exc_info()[2],
                )
    finally:
        _CONTEXT_STACK.pop()


@_decorator
def user_entry_point(wrapped, instance, args, kwargs):
    """
    Decorator for wrapping functions that enter into a user context.  This will help us differentiate user-created
    failures even when it is re-entrant into system code.

    Note: a user_entry_point can ONLY ever be called from within a @system_entry_point wrapped function, therefore,
    we can always ensure we will hit a system_entry_point to correctly reformat our exceptions.  Also, any exception
    we create here will only be handled within our system code so we don't need to worry about leaking weird exceptions
    to the user.
    """
    try:
        _CONTEXT_STACK.append(_USER_CONTEXT)
        if _is_base_context():
            try:
                return wrapped(*args, **kwargs)
            except FlyteScopedException as ex:
                _reraise(ex.type, ex.value, ex.traceback)
        else:
            try:
                return wrapped(*args, **kwargs)
            except FlyteScopedException:
                # Just pass on the already wrapped and scoped exception
                _reraise(*_exc_info())
            except _user_exceptions.FlyteUserException:
                _reraise(
                    FlyteScopedUserException,
                    FlyteScopedUserException(*_exc_info()),
                    _exc_info()[2],
                )
            except _system_exceptions.FlyteSystemException:
                _reraise(
                    FlyteScopedSystemException,
                    FlyteScopedSystemException(*_exc_info()),
                    _exc_info()[2],
                )
            except Exception:
                # Any non-platform raised exception is a user exception.
                # This will also catch FlyteUserException re-raised by the system_entry_point handler
                _reraise(
                    FlyteScopedUserException,
                    FlyteScopedUserException(*_exc_info()),
                    _exc_info()[2],
                )
    finally:
        _CONTEXT_STACK.pop()



from flyteidl.core import errors_pb2 as _errors_pb2

from flytekit.models import common as _common


class ContainerError(_common.FlyteIdlEntity):
    class Kind(object):
        NON_RECOVERABLE = _errors_pb2.ContainerError.NON_RECOVERABLE
        RECOVERABLE = _errors_pb2.ContainerError.RECOVERABLE

    def __init__(self, code, message, kind):
        """
        :param Text code: A succinct code about the error
        :param Text message: Whatever message you want to surface about the error
        :param int kind: A value from the ContainerError.Kind enum.
        """
        self._code = code
        self._message = message
        self._kind = kind

    @property
    def code(self):
        """
        :rtype: Text
        """
        return self._code

    @property
    def message(self):
        """
        :rtype: Text
        """
        return self._message

    @property
    def kind(self):
        """
        :rtype: int
        """
        return self._kind

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.errors_pb2.ContainerError
        """
        return _errors_pb2.ContainerError(code=self.code, message=self.message, kind=self.kind)

    @classmethod
    def from_flyte_idl(cls, proto):
        """
        :param flyteidl.core.errors_pb2.ContainerError proto:
        :rtype: ContainerError
        """
        return cls(proto.code, proto.message, proto.kind)


class ErrorDocument(_common.FlyteIdlEntity):
    def __init__(self, error):
        """
        :param ContainerError error:
        """
        self._error = error

    @property
    def error(self):
        """
        :rtype: ContainerError
        """
        return self._error

    def to_flyte_idl(self):
        """
        :rtype: flyteidl.core.errors_pb2.ErrorDocument
        """
        return _errors_pb2.ErrorDocument(error=self.error.to_flyte_idl())

    @classmethod
    def from_flyte_idl(cls, proto):
        """
        :param flyteidl.core.errors_pb2.ErrorDocument proto:
        :rtype: ErrorDocument
        """
        return cls(ContainerError.from_flyte_idl(proto.error))



import datetime as _datetime
import typing

from flytekit.common.tasks.sagemaker.custom_training_job_task import CustomTrainingJobTask
from flytekit.common.tasks.sagemaker.distributed_training import DefaultOutputPersistPredicate
from flytekit.models.sagemaker import training_job as _training_job_models


def custom_training_job_task(
    _task_function=None,
    algorithm_specification: _training_job_models.AlgorithmSpecification = None,
    training_job_resource_config: _training_job_models.TrainingJobResourceConfig = None,
    cache_version: str = "",
    retries: int = 0,
    deprecated: str = "",
    storage_request: str = None,
    cpu_request: str = None,
    gpu_request: str = None,
    memory_request: str = None,
    storage_limit: str = None,
    cpu_limit: str = None,
    gpu_limit: str = None,
    memory_limit: str = None,
    cache: bool = False,
    timeout: _datetime.timedelta = None,
    environment: typing.Dict[str, str] = None,
    cls: typing.Type = None,
    output_persist_predicate: typing.Callable = DefaultOutputPersistPredicate(),
):
    """
    Decorator to create a Custom Training Job definition.  This task will run as a single unit of work on the platform.

    .. code-block:: python

        @inputs(int_list=[Types.Integer])
        @outputs(sum_of_list=Types.Integer
        @custom_task
        def my_task(wf_params, int_list, sum_of_list):
            sum_of_list.set(sum(int_list))

    :param _task_function: this is the decorated method and shouldn't be declared explicitly.  The function must
        take a first argument, and then named arguments matching those defined in @inputs and @outputs.  No keyword
        arguments are allowed for wrapped task functions.

    :param _training_job_models.AlgorithmSpecification algorithm_specification: This represents the algorithm specification

    :param _training_job_models.TrainingJobResourceConfig training_job_resource_config: This represents the training job config.

    :param Text cache_version: [optional] string representing logical version for discovery.  This field should be
        updated whenever the underlying algorithm changes.

        .. note::

            This argument is required to be a non-empty string if `cache` is True.

    :param int retries: [optional] integer determining number of times task can be retried on
        :py:exc:`flytekit.sdk.exceptions.RecoverableException` or transient platform failures.  Defaults
        to 0.

        .. note::

            If retries > 0, the task must be able to recover from any remote state created within the user code.  It is
            strongly recommended that tasks are written to be idempotent.

    :param Text deprecated: [optional] string that should be provided if this task is deprecated.  The string
        will be logged as a warning so it should contain information regarding how to update to a newer task.

    :param Text storage_request: [optional] Kubernetes resource string for lower-bound of disk storage space
        for the task to run.  Default is set by platform-level configuration.

        .. note::

            This is currently not supported by the platform.

    :param Text cpu_request: [optional] Kubernetes resource string for lower-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text gpu_request: [optional] Kubernetes resource string for lower-bound of desired GPUs.
        Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text memory_request: [optional]  Kubernetes resource string for lower-bound of physical memory
        necessary for the task to execute.  Default is set by platform-level configuration.

        TODO: Add links to resource string documentation for Kubernetes

    :param Text storage_limit: [optional] Kubernetes resource string for upper-bound of disk storage space
        for the task to run.  This amount is not guaranteed!  If not specified, it is set equal to storage_request.

        .. note::

            This is currently not supported by the platform.

    :param Text cpu_limit: [optional] Kubernetes resource string for upper-bound of cores for the task to execute.
        This can be set to a fractional portion of a CPU. This amount is not guaranteed!  If not specified,
        it is set equal to cpu_request.

    :param Text gpu_limit: [optional] Kubernetes resource string for upper-bound of desired GPUs. This amount is not
        guaranteed!  If not specified, it is set equal to gpu_request.

    :param Text memory_limit: [optional]  Kubernetes resource string for upper-bound of physical memory
        necessary for the task to execute.  This amount is not guaranteed!  If not specified, it is set equal to
        memory_request.

    :param bool cache: [optional] boolean describing if the outputs of this task should be cached and
        re-usable.

    :param datetime.timedelta timeout: [optional] describes how long the task should be allowed to
        run at max before triggering a retry (if retries are enabled).  By default, tasks are allowed to run
        indefinitely.  If a null timedelta is passed (i.e. timedelta(seconds=0)), the task will not timeout.

    :param dict[Text,Text] environment: [optional] environment variables to set when executing this task.

    :param cls: This can be used to override the task implementation with a user-defined extension. The class
        provided must be a subclass of flytekit.common.tasks.sdk_runnable.SdkRunnableTask.  A user can use this to
        inject bespoke logic into the base Flyte programming model.

    :param Callable output_persist_predicate: [optional] This callable should return a boolean and is used to indicate whether
        the current copy (i.e., an instance of the task running on a particular node inside the worker pool) would
        write output.

    :rtype: flytekit.common.tasks.sagemaker.custom_training_job_task.CustomTrainingJobTask
    """

    def wrapper(fn):
        return (cls or CustomTrainingJobTask)(
            task_function=fn,
            cache_version=cache_version,
            retries=retries,
            deprecated=deprecated,
            storage_request=storage_request,
            cpu_request=cpu_request,
            gpu_request=gpu_request,
            memory_request=memory_request,
            storage_limit=storage_limit,
            cpu_limit=cpu_limit,
            gpu_limit=gpu_limit,
            memory_limit=memory_limit,
            cache=cache,
            timeout=timeout or _datetime.timedelta(seconds=0),
            environment=environment,
            algorithm_specification=algorithm_specification,
            training_job_resource_config=training_job_resource_config,
            output_persist_predicate=output_persist_predicate,
        )

    if _task_function:
        return wrapper(_task_function)
    else:
        return wrapper



import pytest
from flytekitplugins.awssagemaker.hpo import (
    HPOJob,
    HPOTuningJobConfigTransformer,
    ParameterRangesTransformer,
    SagemakerHPOTask,
)
from flytekitplugins.awssagemaker.training import SagemakerBuiltinAlgorithmsTask, SagemakerTrainingJobConfig

from flytekit import FlyteContext
from flytekit.common.types.primitives import Generic
from flytekit.models.sagemaker.hpo_job import (
    HyperparameterTuningJobConfig,
    HyperparameterTuningObjective,
    HyperparameterTuningObjectiveType,
    TrainingJobEarlyStoppingType,
)
from flytekit.models.sagemaker.parameter_ranges import IntegerParameterRange, ParameterRangeOneOf
from flytekit.models.sagemaker.training_job import AlgorithmName, AlgorithmSpecification, TrainingJobResourceConfig
from plugins.tests.awssagemaker.test_training import _get_reg_settings


def test_hpo_for_builtin():
    trainer = SagemakerBuiltinAlgorithmsTask(
        name="builtin-trainer",
        task_config=SagemakerTrainingJobConfig(
            training_job_resource_config=TrainingJobResourceConfig(
                instance_count=1,
                instance_type="ml-xlarge",
                volume_size_in_gb=1,
            ),
            algorithm_specification=AlgorithmSpecification(
                algorithm_name=AlgorithmName.XGBOOST,
            ),
        ),
    )

    hpo = SagemakerHPOTask(
        name="test",
        task_config=HPOJob(10, 10, ["x"]),
        training_task=trainer,
    )

    assert hpo.python_interface.inputs.keys() == {
        "static_hyperparameters",
        "train",
        "validation",
        "hyperparameter_tuning_job_config",
        "x",
    }
    assert hpo.python_interface.outputs.keys() == {"model"}

    assert hpo.get_custom(_get_reg_settings()) == {
        "maxNumberOfTrainingJobs": "10",
        "maxParallelTrainingJobs": "10",
        "trainingJob": {
            "algorithmSpecification": {"algorithmName": "XGBOOST"},
            "trainingJobResourceConfig": {"instanceCount": "1", "instanceType": "ml-xlarge", "volumeSizeInGb": "1"},
        },
    }

    with pytest.raises(NotImplementedError):
        hpo(
            static_hyperparameters={},
            train="",
            validation="",
            hyperparameter_tuning_job_config=HyperparameterTuningJobConfig(
                tuning_strategy=1,
                tuning_objective=HyperparameterTuningObjective(
                    objective_type=HyperparameterTuningObjectiveType.MINIMIZE,
                    metric_name="x",
                ),
                training_job_early_stopping_type=TrainingJobEarlyStoppingType.OFF,
            ),
            x=ParameterRangeOneOf(param=IntegerParameterRange(10, 1, 1)),
        )


def test_hpoconfig_transformer():
    t = HPOTuningJobConfigTransformer()
    assert t.get_literal_type(HyperparameterTuningJobConfig) == Generic.to_flyte_literal_type()
    o = HyperparameterTuningJobConfig(
        tuning_strategy=1,
        tuning_objective=HyperparameterTuningObjective(
            objective_type=HyperparameterTuningObjectiveType.MINIMIZE,
            metric_name="x",
        ),
        training_job_early_stopping_type=TrainingJobEarlyStoppingType.OFF,
    )
    ctx = FlyteContext.current_context()
    lit = t.to_literal(ctx, python_val=o, python_type=HyperparameterTuningJobConfig, expected=None)
    assert lit is not None
    assert lit.scalar.generic is not None
    ro = t.to_python_value(ctx, lit, HyperparameterTuningJobConfig)
    assert ro is not None
    assert ro == o


def test_parameter_ranges_transformer():
    t = ParameterRangesTransformer()
    assert t.get_literal_type(ParameterRangeOneOf) == Generic.to_flyte_literal_type()
    o = ParameterRangeOneOf(param=IntegerParameterRange(10, 0, 1))
    ctx = FlyteContext.current_context()
    lit = t.to_literal(ctx, python_val=o, python_type=ParameterRangeOneOf, expected=None)
    assert lit is not None
    assert lit.scalar.generic is not None
    ro = t.to_python_value(ctx, lit, ParameterRangeOneOf)
    assert ro is not None
    assert ro == o
